# Reference: https://setuptools.pypa.io/en/latest/build_meta.html#dynamic-build-dependencies-and-other-build-meta-tweaks
from __future__ import annotations

import os
from typing import Any

import setuptools.build_meta
from setuptools.build_meta import *  # noqa: F403


def get_requires_for_build_wheel(config_settings: dict[str, Any] | None = None) -> list[str]:
    base = setuptools.build_meta.get_requires_for_build_wheel(config_settings)

    if os.getenv("BUILD_RUST", "").lower() in {"1", "true"} or (config_settings and "--build-rust" in config_settings):
        return [*base, "setuptools-rust"]

    return base



# References:
# - https://github.com/libyal/libbde
# - https://github.com/Aorimn/dislocker
# - https://github.com/thewhiteninja/ntfstool
# - https://gitlab.com/cryptsetup/cryptsetup
# - fvevol.sys
# - fveapi.dll

from __future__ import annotations

import io
import logging
import os
import struct
from bisect import bisect_right
from operator import itemgetter
from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from dissect.util.stream import AlignedStream

from dissect.fve.bde.c_bde import (
    BITLOCKER_SIGNATURE,
    CIPHER_MAP,
    EOW_INFORMATION_OFFSET_GUID,
    FVE_DATUM_ROLE,
    FVE_DATUM_TYPE,
    FVE_STATE,
    INFORMATION_OFFSET_GUID,
    c_bde,
)
from dissect.fve.bde.eow import EowInformation
from dissect.fve.bde.information import Dataset, Information, KeyDatum
from dissect.fve.bde.key import derive_recovery_key, derive_user_key, stretch
from dissect.fve.crypto import create_cipher
from dissect.fve.exception import InvalidHeaderError

if TYPE_CHECKING:
    from collections.abc import Iterator
    from uuid import UUID

    from dissect.fve.bde.information import VmkInfoDatum

Run = tuple[int, int, int]

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_BDE", "CRITICAL"))


class BDE:
    """Bitlocker disk encryption."""

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.boot_sector = BootSector(fh)

        self._available_information: list[Information] = []
        for offset in self.boot_sector.information_offsets:
            try:
                self._available_information.append(Information(self.fh, offset))
            except InvalidHeaderError as e:  # noqa: PERF203
                log.warning("Failed to parse BDE information at offset 0x%x", offset, exc_info=e)

        self._valid_information = [info for info in self._available_information if info.is_valid()]
        if not self._valid_information:
            raise InvalidHeaderError("No valid BDE information found")
        self.information = self._valid_information[0]

        self.eow_information = None
        self._available_eow_information: list[EowInformation] = []
        for offset in self.boot_sector.eow_offsets:
            try:
                self._available_eow_information.append(EowInformation(self.fh, offset))
            except InvalidHeaderError as e:  # noqa: PERF203
                log.warning("Failed to parse BDE EOW information at offset 0x%x", offset, exc_info=e)

        self._valid_eow_information = [info for info in self._available_eow_information if info.is_valid()]
        if self._available_eow_information and not self._valid_eow_information:
            raise InvalidHeaderError("No valid EOW information found")

        if self._valid_eow_information:
            self.eow_information = self._valid_eow_information[0]

        self._fvek_datum = None
        self._fvek_type = None
        self._fvek = None

    @property
    def identifiers(self) -> list[UUID]:
        datums = self.information.dataset.find_datum(
            role=FVE_DATUM_ROLE.VOLUME_MASTER_KEY_INFO,
            type_=FVE_DATUM_TYPE.VOLUME_MASTER_KEY_INFO,
        )
        return [d.identifier for d in datums]

    @property
    def sector_size(self) -> int:
        return self.boot_sector.sector_size

    @property
    def version(self) -> int:
        return self.information.version

    @property
    def paused(self) -> bool:
        return self.information.current_state == FVE_STATE.PAUSED

    @property
    def decrypted(self) -> bool:
        return self.information.current_state == FVE_STATE.DECRYPTED

    @property
    def encrypted(self) -> bool:
        return not self.decrypted

    @property
    def switching(self) -> bool:
        return self.information.current_state not in (FVE_STATE.DECRYPTED, FVE_STATE.ENCRYPTED)

    @property
    def unlocked(self) -> bool:
        return self._fvek is not None or self.information.current_state == FVE_STATE.DECRYPTED

    def description(self) -> str | None:
        """Return the volume description, if present."""
        return self.information.dataset.find_description()

    def has_clear_key(self) -> bool:
        """Return whether this volume has a clear/obfuscated encryption key. Used in paused volumes."""
        return self.information.dataset.find_clear_vmk() is not None

    def has_recovery_password(self) -> bool:
        """Return whether this volume can be unlocked with a recovery password."""
        return len(list(self.information.dataset.find_recovery_vmk())) != 0

    def has_passphrase(self) -> bool:
        """Return whether this volume can be unlocked with a user passphrase."""
        return len(list(self.information.dataset.find_passphrase_vmk())) != 0

    def has_bek(self) -> bool:
        """Return whether this volume can be unlocked with a BEK file."""
        return len(list(self.information.dataset.find_external_vmk())) != 0

    def unlock(self, key: bytes) -> BDE:
        """Unlock this volume with the specified encryption key."""
        self.information.check_integrity(key)

        fvek = self.information.dataset.find_fvek()
        if not fvek:
            raise ValueError("No FVEK found")

        fvek = fvek.unbox(key)
        if not isinstance(fvek, KeyDatum):
            raise TypeError("Invalid unboxed FVEK")

        self._fvek_datum = fvek
        self._fvek_type = fvek.key_type
        self._fvek = fvek.data

        return self

    def unlock_with_clear_key(self) -> BDE:
        """Unlock this volume with the clear/obfuscated key."""
        if not (vmk := self.information.dataset.find_clear_vmk()):
            raise ValueError("No clear VMK found")

        return self.unlock(vmk.decrypt(vmk.clear_key()))

    def unlock_with_recovery_password(self, recovery_password: str, identifier: UUID | str | None = None) -> BDE:
        """Unlock this volume with the recovery password."""
        recovery_key = derive_recovery_key(recovery_password)
        try:
            return self._unlock_with_user_key(self.information.dataset.find_recovery_vmk(), recovery_key, identifier)
        except ValueError as e:
            raise ValueError("Unable to unlock with given recovery password") from e

    def unlock_with_passphrase(self, passphrase: str, identifier: UUID | str | None = None) -> BDE:
        """Unlock this volume with the user passphrase."""
        user_key = derive_user_key(passphrase)
        try:
            return self._unlock_with_user_key(self.information.dataset.find_passphrase_vmk(), user_key, identifier)
        except ValueError as e:
            raise ValueError("Unable to unlock with given passphrase") from e

    def unlock_with_bek(self, bek_fh: BinaryIO) -> BDE:
        """Unlock this volume with a BEK file."""
        bek_ds = Dataset(bek_fh)
        startup_key = bek_ds.find_startup_key()
        if not startup_key:
            raise ValueError("Unable to unlock with BEK, no startup key found")

        for vmk in self.information.dataset.find_external_vmk():
            if vmk.identifier == startup_key.identifier:
                break
        else:
            raise ValueError("Unable to unlock with BEK, no matching VMK found")

        decrypted_key = vmk.decrypt(startup_key.external_key())
        return self.unlock(decrypted_key)

    def unlock_with_fvek(self, key: bytes) -> BDE:
        """Unlock this volume with a raw FVEK key."""
        self._fvek_type = self.information.dataset.fvek_type
        self._fvek = key
        return self

    def _unlock_with_user_key(
        self, vmks: list[VmkInfoDatum], user_key: bytes, identifier: UUID | str | None = None
    ) -> BDE:
        decrypted_key = None
        for vmk in vmks:
            if identifier and str(identifier) != str(vmk.identifier):
                continue

            # There should only be one stretch key
            stretch_key = vmk.stretch_key(None)
            if not stretch_key:
                continue  # Shouldn't happen

            aes_key = stretch(user_key, stretch_key.salt)
            try:
                decrypted_key = vmk.decrypt(aes_key)
                break
            except ValueError:
                continue
        else:
            raise ValueError("No compatible VMK found, incorrect key or identifier")

        return self.unlock(decrypted_key)

    def open(self) -> BitlockerStream:
        """Open this volume and return a readable (decrypted) stream."""
        if not self.unlocked:
            raise ValueError("Volume is locked")
        return BitlockerStream(self)

    def reserved_regions(self) -> list[tuple[int, int]]:
        """Return a list of reserved regions for this volume.

        Some areas of the volume must "fake" return all null bytes when read.
        This includes things like the information regions.

        Reference:
        - InitializeFilterData
        - FveLibIdentifyCurrentRegionTypeAndEnd
        """
        regions = []

        if self.version == 1:
            information_size = (self.boot_sector.cluster_size + 0x3FFF) & ~(self.boot_sector.cluster_size - 1)
        elif self.version >= 2:
            information_size = ~(self.sector_size - 1) & (self.sector_size + 0xFFFF)

        # All information offsets are reserved regions
        regions.extend(
            [
                (offset // self.sector_size, information_size // self.sector_size)
                for offset in self.information.information_offset
            ]
        )

        if self.version >= 2:
            num_sectors = self.information.virtualized_sectors or 1
            regions.append((self.information.virtualized_block_offset // self.sector_size, num_sectors))

        for eow_info in self._valid_eow_information:
            eow_information_size = ~(self.sector_size - 1) & (eow_info.size + self.sector_size - 1)
            regions.append((eow_info.offset // self.sector_size, eow_information_size // self.sector_size))

            for bitmap in eow_info.bitmaps:
                regions.append((bitmap.offset // self.sector_size, bitmap.size // self.sector_size))
                regions.append((bitmap.conv_log_offset // self.sector_size, eow_info.conv_log_size // self.sector_size))

        # In progress encryption/decryption with dirty state
        if self.information.current_state == FVE_STATE.SWITCHING_DIRTY and self.information.state_size:
            regions.append(
                (
                    self.information.state_offset // self.sector_size,
                    self.information.state_size // self.sector_size,
                )
            )

        return sorted(set(regions), key=itemgetter(0))


class BootSector:
    """Bitlocker boot sector parsing.

    Bitlocker seems to do some funny stuff with the boot sector. Instead of trying to make sense of that,
    just do what Microsoft does in their driver: looking for specific GUIDs to determine the version.

    If no GUIDs can be found, but the Oem string still says -FVE-FS-, we're dealing with a legacy Vista volume.
    """

    def __init__(self, fh: BinaryIO):
        buf = fh.read(512)

        self.boot_sector = c_bde.BOOT_SECTOR(buf)
        self.sector_size = self.boot_sector.Bpb.BytesPerSector
        self.cluster_size = self.sector_size * self.boot_sector.Bpb.SectorsPerCluster

        self.guid = None
        self.information_offsets = []
        self.eow_offsets = []

        info_guid_offset = buf.find(INFORMATION_OFFSET_GUID.bytes_le)
        eow_guid_offset = buf.find(EOW_INFORMATION_OFFSET_GUID.bytes_le)

        if eow_guid_offset != -1:
            info = c_bde.FVE_EOW_GUID_RECOGNITION(buf[eow_guid_offset:])
            self.guid = EOW_INFORMATION_OFFSET_GUID
            self.information_offsets = info.InformationOffset
            self.eow_offsets = info.EowOffset
        elif info_guid_offset != -1:
            info = c_bde.FVE_GUID_RECOGNITION(buf[info_guid_offset:])
            self.guid = INFORMATION_OFFSET_GUID
            self.information_offsets = info.InformationOffset
        elif self.boot_sector.Oem == BITLOCKER_SIGNATURE:
            self.information_offsets = [self.boot_sector.InformationLcn * self.cluster_size]
        else:
            raise ValueError("Not a BDE volume")


class BitlockerStream(AlignedStream):
    """Transparently decrypting Bitlocker stream.

    Provides a transparently decrypted Bitlocker stream for reading. Takes care of the reserved regions, as well
    as the virtualized blocks in Vista and newer Bitlocker versions.

    For Vista, the first 0x2000 bytes aren't actually encrypted. The very first sector is obviously modified to
    contain the Bitlocker information, so when reading that sector we must patch the Oem ID to be the NTFS one,
    as well as replacing the secondary MFT location with one that's located in the Information structure.

    For newer versions, the first N sectors (usually 16) _are_ encrypted, but have been placed elsewhere on the
    volume. The location and amount of so-called virtualized sectors are specified in the Information structure.

    The Microsoft implementation works on a byte level, for the time being it's easier for us to work on sector
    level. I haven't seen a reason why this would break, yet.
    """

    RUN_PLAIN = 0
    RUN_VISTA_HEADER = 1
    RUN_SPARSE = 2
    RUN_ENCRYPTED = 3

    def __init__(self, bde: BDE):
        self.bde = bde
        self._fh = bde.fh

        size = getattr(bde.fh, "size", None)
        try:
            if size is None:
                bde.fh.seek(0, io.SEEK_END)
                size = bde.fh.tell()
        except Exception:
            pass

        self.sector_size = bde.sector_size
        if self.bde.encrypted:
            self.encrypted = True
            self.cipher = create_cipher(
                CIPHER_MAP[bde._fvek_type],
                bde._fvek,
                sector_size=self.sector_size,
                iv_sector_size=self.sector_size,
            )
        else:
            self.encrypted = False
            self.cipher = None

        self._reserved_regions = bde.reserved_regions()
        self._virtualized_block_offset = bde.information.virtualized_block_offset
        self._virtualized_block_sector = self._virtualized_block_offset // self.sector_size
        self._virtualized_sector_count = bde.information.virtualized_sectors

        self._state_offset_sector = bde.information.state_offset // self.sector_size

        self.is_eow = bde.eow_information is not None
        self._eow_bitmaps = []
        self._eow_bitmap_lookup = []
        self._eow_sectors_per_chunk = None

        if self.is_eow:
            self._eow_bitmaps = bde.eow_information.bitmaps
            self._eow_bitmap_lookup = [bm.region_offset // self.sector_size for bm in self._eow_bitmaps]
            self._eow_sectors_per_chunk = self.bde.eow_information.chunk_size // self.sector_size

        super().__init__(size=size)

    def _iter_run_state(self, sector: int, count: int) -> Iterator[Run]:
        while count > 0:
            if self.is_eow:
                bitmap_idx = bisect_right(self._eow_bitmap_lookup, sector)
                bitmap = self._eow_bitmaps[bitmap_idx - 1]

                relative_sector = sector - (bitmap.region_offset // self.sector_size)
                chunk = (relative_sector * self.sector_size) // self.bde.eow_information.chunk_size
                chunk_count = -(-count // self._eow_sectors_per_chunk)

                for bit_set, bit_count in bitmap.runs(chunk, chunk_count):
                    run_type = BitlockerStream.RUN_ENCRYPTED if bit_set else BitlockerStream.RUN_PLAIN
                    run_count = min(count, bit_count * self._eow_sectors_per_chunk)

                    yield (run_type, sector, run_count)

                    sector += run_count
                    count -= run_count
            else:
                # The StateOffset determines how much of the volume is encrypted.
                # In pre-EOW Bitlocker, it's actually used to determine how much of the volume is encrypted
                # for partially encrypted volumes, but since EOW, it seems to just contain the volume size.
                # Pre-EOW volumes are encrypted back to front, so reading beyond the StateOffset means reading
                # plaintext data.
                if self._state_offset_sector and sector < self._state_offset_sector:
                    remaining_sectors = min(self._state_offset_sector - sector, count)
                    yield (BitlockerStream.RUN_ENCRYPTED, sector, remaining_sectors)

                    sector += remaining_sectors
                    count -= remaining_sectors

                if self._state_offset_sector and sector >= self._state_offset_sector:
                    yield (BitlockerStream.RUN_PLAIN, sector, count)
                elif self.encrypted:
                    yield (BitlockerStream.RUN_ENCRYPTED, sector, count)
                else:
                    yield (BitlockerStream.RUN_PLAIN, sector, count)

                sector += count
                count -= count

    def _iter_runs(self, offset: int, length: int) -> Iterator[Run]:
        sector = offset // self.sector_size
        count = -(-length // self.sector_size)

        while count != 0:
            # Vista volume header behaviour
            if self.bde.version == 1 and sector < 0x2000 // self.sector_size:
                if sector == 0:
                    yield (BitlockerStream.RUN_VISTA_HEADER, sector, 1)

                    sector += 1
                    count -= 1

                # Intentionally fall through
                if count:
                    remaining_sectors = min((0x2000 // self.sector_size) - sector, count)

                    yield (BitlockerStream.RUN_PLAIN, sector, remaining_sectors)

                    sector += remaining_sectors
                    count -= remaining_sectors

            # Only on Bitlocker version >= 2
            if sector < self._virtualized_sector_count:
                remaining_sectors = min(self._virtualized_sector_count - sector, count)

                yield from self._iter_run_state(sector + self._virtualized_block_sector, remaining_sectors)

                sector += remaining_sectors
                count -= remaining_sectors

            for region_start, region_size in self._reserved_regions:
                if count == 0:
                    break

                region_end = region_start + region_size

                # Starts outside a region but ends in or after it
                if sector < region_start < sector + count:
                    remaining_sectors = min(region_start - sector, count)

                    yield from self._iter_run_state(sector, remaining_sectors)

                    sector += remaining_sectors
                    count -= remaining_sectors

                # Starts in a region
                if region_start <= sector < region_end:
                    remaining_sectors = min(region_end - sector, count)

                    yield (BitlockerStream.RUN_SPARSE, sector, remaining_sectors)

                    sector += remaining_sectors
                    count -= remaining_sectors
            else:
                yield from self._iter_run_state(sector, count)

                sector += count
                count -= count

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        for run_type, read_sector, sector_count in _consolidate_runs(self._iter_runs(offset, length)):
            if run_type == BitlockerStream.RUN_PLAIN:
                self._fh.seek(read_sector * self.sector_size)
                result.append(self._fh.read(sector_count * self.sector_size))
            elif run_type == BitlockerStream.RUN_VISTA_HEADER:
                self._fh.seek(read_sector * self.sector_size)
                buf = bytearray(self._fh.read(sector_count * self.sector_size))
                buf[0x03:0x0B] = b"NTFS    "
                buf[0x38:0x40] = struct.pack("<Q", self.bde.information.header.Mft2StartLcn)
                result.append(bytes(buf))
            elif run_type == BitlockerStream.RUN_SPARSE:
                result.append(b"\x00" * self.sector_size * sector_count)
            elif run_type == BitlockerStream.RUN_ENCRYPTED:
                self._fh.seek(read_sector * self.sector_size)
                buf = self._fh.read(sector_count * self.sector_size)
                result.append(self.cipher.decrypt(buf, read_sector))

        return b"".join(result)


def _consolidate_runs(it: Iterator[Run]) -> Iterator[Run]:
    current_type = None
    current_sector = None
    current_count = 0

    for run_type, sector, count in it:
        if current_type is None:
            current_type = run_type
            current_sector = sector
            current_count = count
            continue

        if current_type != run_type or current_sector + current_count != sector:
            yield (current_type, current_sector, current_count)

            current_type = run_type
            current_sector = sector
            current_count = count
        else:
            current_count += count

    if current_type is not None:
        yield (current_type, current_sector, current_count)


def is_bde_volume(fh: BinaryIO) -> bool:
    stored_position = fh.tell()
    try:
        fh.seek(0)
        BootSector(fh)
    except ValueError:
        return False
    else:
        return True
    finally:
        fh.seek(stored_position)



from __future__ import annotations

import argparse
import io
import sys
import traceback
from pathlib import Path
from typing import BinaryIO

from dissect.target import container, volume

from dissect.fve.bde import BDE, is_bde_volume
from dissect.fve.crypto.dmcrypt import CryptStream
from dissect.fve.luks import LUKS, is_luks_volume

try:
    from rich.progress import (
        BarColumn,
        DownloadColumn,
        Progress,
        TextColumn,
        TimeRemainingColumn,
        TransferSpeedColumn,
    )

    progress = Progress(
        TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
        BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
        transient=True,
    )

    log = progress.console.log
except ImportError:

    class Progress:
        def __init__(self):
            self.filename = None
            self.total = None

            self.position = 0

        def __enter__(self) -> None:
            pass

        def __exit__(self, *args, **kwargs) -> None:
            sys.stderr.write("\n")
            sys.stderr.flush()

        def add_task(self, name: str, filename: str, total: int, **kwargs) -> None:
            self.filename = filename
            self.total = total

        def update(self, task_id: int, advance: int) -> None:
            self.position += advance

            sys.stderr.write(f"\r{self.filename} {(self.position / self.total) * 100:0.2f}%")
            sys.stderr.flush()

    import logging

    progress = Progress()

    logger = logging.getLogger(__name__)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter("%(message)s"))
    stream_handler.setLevel(logging.INFO)
    logger.addHandler(stream_handler)
    logger.setLevel(logging.INFO)

    log = logger.info


def stream(
    fhin: BinaryIO,
    fhout: BinaryIO,
    offset: int,
    length: int,
    chunk_size: int = io.DEFAULT_BUFFER_SIZE,
    task_id: int | None = None,
) -> None:
    fhin.seek(offset)
    while length != 0:
        read_size = min(length, chunk_size)
        fhout.write(fhin.read(read_size))

        progress.update(task_id, advance=read_size)

        length -= read_size


def open_fve(vol: BinaryIO, args: argparse.Namespace) -> BinaryIO:
    # Currently only BDE and LUKS
    if is_bde_volume(vol):
        return _open_bde(vol, args)

    if is_luks_volume(vol):
        return _open_luks(vol, args)

    # Plain volume, return itself
    return vol


def _open_bde(vol: BinaryIO, args: argparse.Namespace) -> BinaryIO | None:
    bde = BDE(vol)

    if bde.has_clear_key():
        bde.unlock_with_clear_key()
    else:
        if args.passphrase and bde.has_passphrase():
            try:
                bde.unlock_with_passphrase(args.passphrase)
                log("Unlocked BDE volume with passphrase")
            except Exception as e:
                log(f"Failed to unlock BDE volume with passphrase: {e}")

        elif args.recovery and bde.has_recovery_password():
            try:
                bde.unlock_with_recovery_password(args.recovery)
                log("Unlocked BDE volume with recovery password")
            except Exception as e:
                log(f"Failed to unlock BDE volume with recovery password: {e}")

        elif args.unlock_file:
            try:
                with args.unlock_file.open("rb") as fh:
                    bde.unlock_with_bek(fh)
                log("Unlocked BDE volume with BEK")
            except Exception as e:
                log(f"Failed to unlock BDE volume with BEK: {e}")

    if not bde.unlocked:
        log("Failed to unlock BDE volume")
        return None

    return bde.open()


def _open_luks(vol: BinaryIO, args: argparse.Namespace) -> BinaryIO | None:
    luks = LUKS(vol)

    if args.passphrase:
        try:
            luks.unlock_with_passphrase(args.passphrase, args.key_slot)
            log("Unlocked LUKS volume with passphrase")
        except Exception as e:
            log(f"Failed to unlock LUKS volume with passphrase: {e}")
    elif args.unlock_file:
        try:
            luks.unlock_with_key_file(args.unlock_file, args.keyfile_offset, args.keyfile_size, args.key_slot)
            log("Unlocked LUKS volume with key file")
        except Exception as e:
            log(f"Failed to unlock LUKS volume with key file: {e}")

    if not luks.unlocked:
        log("Failed to unlock LUKS volume")
        return None

    return luks.open()


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=Path, help="path to container with encrypted volume")
    parser.add_argument("-p", "--passphrase", type=str, help="user passphrase")
    parser.add_argument("-r", "--recovery", type=str, help="recovery passphrase")
    parser.add_argument("-f", "--unlock-file", type=Path, help="unlock file")
    parser.add_argument("--key-slot", type=int, help="LUKS keyslot")
    parser.add_argument("--keyfile-offset", type=int, help="LUKS keyfile offset")
    parser.add_argument("--keyfile-size", type=int, help="LUKS keyfile size")
    parser.add_argument("-o", "--output", type=Path, required=True, help="path to output file")
    parser.add_argument("-v", "--verbose", action="count", default=3, help="increase output verbosity")
    args = parser.parse_args()

    in_path = args.input.resolve()

    if not in_path.exists():
        parser.exit(f"Input file doesn't exist: {in_path}")

    disk = container.open(in_path)
    try:
        vs = volume.open(disk)
        disk_volumes = vs.volumes
    except Exception:
        log("Container has no volume system, treating as raw instead")
        disk_volumes = [volume.Volume(disk, 1, 0, disk.size, None, None, disk=disk)]

    volumes = []
    for vol in disk_volumes:
        fve_vol = None

        try:
            fve_vol = open_fve(vol, args)
        except Exception:
            log(traceback.format_exc())
            log("Exception opening FVE volume")

        if fve_vol is None:
            parser.exit(f"Failed to open FVE volume: {vol}")
        else:
            volumes.append((vol, fve_vol))

    task_id = progress.add_task("decrypt", start=True, visible=True, filename=in_path.name, total=disk.size)

    offset = 0
    with progress, args.output.open("wb") as fh:
        for vol, fve_vol in volumes:
            if offset != vol.offset:
                # We're not to the beginning of the volume yet, fill in
                stream(disk, fh, offset, vol.offset - offset, task_id=task_id)
                offset = vol.offset

            # Stream the decrypted volume
            src_vol = fve_vol or vol
            stream(src_vol, fh, 0, src_vol.size, task_id=task_id)
            offset += src_vol.size

            if isinstance(fve_vol, CryptStream):
                # LUKS volumes don't actually start at the beginning like Bitlocker
                offset += fve_vol.offset

        # There's data after the volumes until the end of the disk
        if offset != disk.size:
            stream(disk, fh, offset, disk.size - offset, task_id=task_id)

    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass



from __future__ import annotations

from dissect.cstruct import cstruct

luks_def = """
/* =========== LUKS1 =========== */
#define LUKS_CIPHERNAME_L       32
#define LUKS_CIPHERMODE_L       32
#define LUKS_HASHSPEC_L         32
#define LUKS_DIGESTSIZE         20  // since SHA1
#define LUKS_HMACSIZE           32
#define LUKS_SALTSIZE           32
#define LUKS_NUMKEYS            8

// Minimal number of iterations
#define LUKS_MKD_ITERATIONS_MIN     1000
#define LUKS_SLOT_ITERATIONS_MIN    1000

// Iteration time for digest in ms
#define LUKS_MKD_ITERATIONS_MS  125

#define LUKS_KEY_DISABLED_OLD   0
#define LUKS_KEY_ENABLED_OLD    0xCAFE

#define LUKS_KEY_DISABLED       0x0000DEAD
#define LUKS_KEY_ENABLED        0x00AC71F3

#define LUKS_STRIPES 4000

// partition header starts with magic
#define LUKS_MAGIC_L            6

/* Actually we need only 37, but we don't want struct autoaligning to kick in */
#define UUID_STRING_L           40

/* Offset to keyslot area [in bytes] */
#define LUKS_ALIGN_KEYSLOTS     4096

/* Maximal LUKS header size, for wipe [in bytes] */
#define LUKS_MAX_KEYSLOT_SIZE   0x1000000   /* 16 MB, up to 32768 bits key */

/* Any integer values are stored in network byte order on disk and must be converted */

/* DISSECT: Keyblock structure currently separated out due to a cstruct limitation */
struct luks_keyblock {
    uint32_t    active;

    /* parameters used for password processing */
    uint32_t    passwordIterations;

    char        passwordSalt[LUKS_SALTSIZE];
    /* parameters used for AF store/load */
    uint32_t    keyMaterialOffset;
    uint32_t    stripes;
};

struct luks_phdr {
    char        magic[LUKS_MAGIC_L];
    uint16_t    version;
    char        cipherName[LUKS_CIPHERNAME_L];
    char        cipherMode[LUKS_CIPHERMODE_L];
    char        hashSpec[LUKS_HASHSPEC_L];
    uint32_t    payloadOffset;
    uint32_t    keyBytes;
    char        mkDigest[LUKS_DIGESTSIZE];
    char        mkDigestSalt[LUKS_SALTSIZE];
    uint32_t    mkDigestIterations;
    char        uuid[UUID_STRING_L];

    /* DISSECT: Keyblock structure currently separated out due to a cstruct limitation */
    luks_keyblock keyblock[LUKS_NUMKEYS];

    /* Align it to 512 sector size */
    char        _padding[432];
};

/* =========== LUKS2 =========== */

#define LUKS2_MAGIC_L           6
#define LUKS2_UUID_L            40
#define LUKS2_LABEL_L           48
#define LUKS2_SALT_L            64
#define LUKS2_CHECKSUM_ALG_L    32
#define LUKS2_CHECKSUM_L        64

#define LUKS2_KEYSLOTS_MAX      32
#define LUKS2_TOKENS_MAX        32
#define LUKS2_SEGMENT_MAX       32

/*
 * LUKS2 header on-disk.
 *
 * Binary header is followed by JSON area.
 * JSON area is followed by keyslot area and data area,
 * these are described in JSON metadata.
 *
 * Note: uuid, csum_alg are intentionally on the same offset as LUKS1
 * (checksum alg replaces hash in LUKS1)
 *
 * String (char) should be zero terminated.
 * Padding should be wiped.
 * Checksum is calculated with csum zeroed (+ full JSON area).
 */
struct luks2_hdr_disk {
    char        magic[LUKS2_MAGIC_L];
    uint16_t    version;                            /* Version 2 */
    uint64_t    hdr_size;                           /* in bytes, including JSON area */
    uint64_t    seqid;                              /* increased on every update */
    char        label[LUKS2_LABEL_L];
    char        checksum_alg[LUKS2_CHECKSUM_ALG_L];
    uint8_t     salt[LUKS2_SALT_L];                 /* unique for every header/offset */
    char        uuid[LUKS2_UUID_L];
    char        subsystem[LUKS2_LABEL_L];           /* owner subsystem label */
    uint64_t    hdr_offset;                         /* offset from device start in bytes */
    char        _padding[184];
    uint8_t     csum[LUKS2_CHECKSUM_L];
    char        _padding4096[7*512];
    /* JSON area starts here */
};
"""

c_luks = cstruct(endian=">").load(luks_def)

LUKS_MAGIC = b"LUKS\xba\xbe"
LUKS2_MAGIC_1ST = b"LUKS\xba\xbe"
LUKS2_MAGIC_2ND = b"SKUL\xba\xbe"

SECONDARY_HEADER_OFFSETS = [
    0x00004000,
    0x00008000,
    0x00010000,
    0x00020000,
    0x00040000,
    0x00080000,
    0x00100000,
    0x00200000,
    0x00400000,
]



from __future__ import annotations

from dissect.util.xmemoryview import xmemoryview


def diffuser_a_decrypt(buffer: memoryview, sector_size: int) -> None:
    a_cycles = 5
    r_a = [9, 0, 13, 0]
    int_size = sector_size >> 2

    buffer_i = xmemoryview(buffer, "<I")

    for _ in range(a_cycles):
        for i in range(int_size):
            buffer_i[i] = (buffer_i[i] + (buffer_i[i - 2] ^ _rotate_left(buffer_i[i - 5], r_a[i % 4]))) & 0xFFFFFFFF


def diffuser_a_encrypt(buffer: memoryview, sector_size: int) -> None:
    a_cycles = 5
    r_a = [9, 0, 13, 0]
    int_size = sector_size >> 2

    buffer_i = xmemoryview(buffer, "<I")

    for _ in range(a_cycles):
        for i in range(int_size - 1, -1, -1):
            buffer_i[i] = (buffer_i[i] - (buffer_i[i - 2] ^ _rotate_left(buffer_i[i - 5], r_a[i % 4]))) & 0xFFFFFFFF


def diffuser_b_decrypt(buffer: memoryview, sector_size: int) -> None:
    b_cycles = 3
    r_b = [0, 10, 0, 25]
    int_size = sector_size >> 2

    buffer_i = xmemoryview(buffer, "<I")

    for _ in range(b_cycles):
        for i in range(int_size):
            buffer_i[i] = (
                buffer_i[i] + (buffer_i[(i + 2) % int_size] ^ _rotate_left(buffer_i[(i + 5) % int_size], r_b[i % 4]))
            ) & 0xFFFFFFFF


def diffuser_b_encrypt(buffer: memoryview, sector_size: int) -> None:
    b_cycles = 3
    r_b = [0, 10, 0, 25]
    int_size = sector_size >> 2

    buffer_i = xmemoryview(buffer, "<I")

    for _ in range(b_cycles):
        for i in range(int_size - 1, -1, -1):
            buffer_i[i] = (
                buffer_i[i] - (buffer_i[(i + 2) % int_size] ^ _rotate_left(buffer_i[(i + 5) % int_size], r_b[i % 4]))
            ) & 0xFFFFFFFF


def _rotate_left(num: int, count: int) -> int:
    return ((num << count) | (num >> (32 - count))) & ((0b1 << 32) - 1)



from __future__ import annotations

import hashlib

from dissect.fve.crypto.util import xor

DIGEST_SIZE = {
    "sha1": 20,
    "sha256": 32,
    "sha512": 64,
}


def _hash(buf: bytes, hash: str, iv: int) -> bytes:
    ctx = hashlib.new(hash)
    ctx.update((iv & 0xFFFFFFFF).to_bytes(4, "big"))
    ctx.update(buf)
    return ctx.digest()


def diffuse(buf: bytes, hash: str) -> bytes:
    buf_size = len(buf)
    digest_size = DIGEST_SIZE[hash]

    view = memoryview(buf)
    result = bytearray(buf_size)

    passes, remainder = divmod(buf_size, digest_size)

    for i in range(passes):
        result[i * digest_size : (i + 1) * digest_size] = _hash(view[i * digest_size : (i + 1) * digest_size], hash, i)

    if remainder:
        result[passes * digest_size : buf_size] = _hash(view[buf_size - remainder : buf_size], hash, passes)[:remainder]

    return bytes(result)


def merge(buf: bytes, block_size: int, block_num: int, hash: str) -> bytes:
    if block_size * block_num > len(buf):
        raise ValueError(f"Unexpected input buffer size ({block_size} * {block_num} != {len(buf)})")

    tmp = bytearray(block_size)
    view = memoryview(buf)

    for i in range(block_num - 1):
        block = view[i * block_size : (i + 1) * block_size]
        xor(block, tmp, output=tmp)
        tmp[:] = diffuse(tmp, hash)

    xor(view[(block_num - 1) * block_size : block_num * block_size], tmp, output=tmp)
    return bytes(tmp)



from __future__ import annotations

from uuid import UUID

from dissect.cstruct import cstruct

bde_def = """
/* ======== Volume header information ======== */

typedef struct _FVE_GUID_RECOGNITION {
    CHAR        Guid[16];
    QWORD       InformationOffset[3];
} FVE_GUID_RECOGNITION;

typedef struct _FVE_EOW_GUID_RECOGNITION {
    CHAR        Guid[16];
    QWORD       InformationOffset[3];
    QWORD       EowOffset[2];
} FVE_EOW_GUID_RECOGNITION;

typedef struct _BIOS_PARAMETER_BLOCK {
    USHORT      BytesPerSector;
    UCHAR       SectorsPerCluster;
    USHORT      ReservedSectors;
    UCHAR       Fats;
    USHORT      RootEntries;
    USHORT      Sectors;
    UCHAR       Media;
    USHORT      SectorsPerFat;
    USHORT      SectorsPerTrack;
    USHORT      Heads;
    ULONG       HiddenSectors;
    ULONG       LargeSectors;
} BIOS_PARAMETER_BLOCK;

typedef struct _BOOT_SECTOR {
    CHAR        Jump[3];
    CHAR        Oem[8];
    BIOS_PARAMETER_BLOCK    Bpb;
    CHAR        Unused0[20];
    union {
        ULONG64 InformationLcn;
        ULONG64 Mft2StartLcn;
    };
    CHAR        Unused1[8];
    ULONG64     PartitionLength;
    CHAR        Unused2[28];
    UCHAR       BytesPerSectorShift;
    UCHAR       SectorsPerClusterShift;
    CHAR        Unused3[402];
} BOOT_SECTOR;

/* ======== FVE information and dataset ======== */

enum FVE_STATE {
    DECRYPTED                   = 1,        /* Decrypted state */
    SWITCHING_DIRTY             = 2,        /* In-progress encryption or decryption of large volumes */
                                            /* StateSize will be non-zero, and there will be a conversion log */
    PAUSED                      = 3,        /* Seen on Vista volume with paused encryption/decryption */
    ENCRYPTED                   = 4,        /* The most common state */
    SWITCHING                   = 5,        /* In-progress encryption or decryption of small volumes */
                                            /* Seen when detaching VHD during encryption/decryption of small disks */
};

enum FVE_KEY_TYPE {
    NONE                        = 0x0000,
    EXTERNAL                    = 0x0005,   /* External VMKs have a USE_KEY with this key type */

    STRETCH_KEY                 = 0x1000,
    STRETCH_KEY_1               = 0x1001,
    AES_CCM_256_0               = 0x2000,
    AES_CCM_256_1               = 0x2001,
    EXTERN_KEY                  = 0x2002,
    VMK                         = 0x2003,
    AES_CCM_256_2               = 0x2004,
    HASH_256                    = 0x2005,

    AES_128_DIFFUSER            = 0x8000,
    AES_256_DIFFUSER            = 0x8001,
    AES_128                     = 0x8002,
    AES_256                     = 0x8003,
    AES_XTS_128                 = 0x8004,
    AES_XTS_256                 = 0x8005,
};

flag FVE_KEY_PROTECTOR {
    CLEAR                       = 0x0000,   /* Also known as "obfuscated" */
    TPM                         = 0x0100,
    EXTERNAL                    = 0x0200,   /* Startup key */
    TPM_PIN                     = 0x0400,
    RECOVERY_PASSWORD           = 0x0800,   /* Recovery password */
    PASSPHRASE                  = 0x2000,   /* User passphrase */
};

flag FVE_KEY_FLAG {
    NONE                        = 0x00,
    ENHANCED_PIN                = 0x04,
    ENHANCED_CRYPTO             = 0x10,
    PBKDF2                      = 0x40,
};

enum FVE_DATUM_ROLE : USHORT {
    PROPERTY                    = 0x0000,

    UNKNOWN_1                   = 0x0001,

    VOLUME_MASTER_KEY_INFO      = 0x0002,
    FULL_VOLUME_ENCRYPTION_KEY  = 0x0003,
    VALIDATION                  = 0x0004,

    UNKNOWN_5                   = 0x0005,

    STARTUP_KEY                 = 0x0006,
    DESCRIPTION                 = 0x0007,

    UNKNOWN_8                   = 0x0008,
    UNKNOWN_9                   = 0x0009,
    UNKNOWN_A                   = 0x000A,
    AUTO_UNLOCK                 = 0x000B,
    FULL_VOLUME_ENCRYPTION_KEY_2    = 0x000C,
    UNKNOWN_D                   = 0x000D,
    UNKNOWN_E                   = 0x000E,

    VIRTUALIZATION_INFO         = 0x000F,
    VALIDATION_HASH             = 0x0011,
};

enum FVE_DATUM_TYPE : USHORT {
    ERASED                      = 0x0000,
    KEY                         = 0x0001,
    UNICODE                     = 0x0002,
    STRETCH_KEY                 = 0x0003,
    USE_KEY                     = 0x0004,
    AES_CCM_ENCRYPTED_KEY       = 0x0005,
    TPM_ENCRYPTED_BLOB          = 0x0006,
    VALIDATION_INFO             = 0x0007,
    VOLUME_MASTER_KEY_INFO      = 0x0008,
    EXTERNAL_INFO               = 0x0009,
    UPDATE                      = 0x000A,
    ERROR_LOG                   = 0x000B,
    ASYMMETRIC_ENCRYPTED_KEY    = 0x000C,
    EXPORTED_KEY                = 0x000D,
    PUBLIC_KEY_INFO             = 0x000E,
    VIRTUALIZATION_INFO         = 0x000F,
    SIMPLE_1                    = 0x0010,
    SIMPLE_2                    = 0x0011,
    CONCAT_HASH_KEY             = 0x0012,
    SIMPLE_3                    = 0x0013,
    SIMPLE_LARGE                = 0x0014,
    BACKUP_INFO                 = 0x0015,
};

typedef struct _FVE_INFORMATION {
    CHAR        Signature[8];
    USHORT      HeaderSize;
    USHORT      Version;
    USHORT      CurrentState;
    USHORT      NextState;
    ULONG64     StateOffset;
    ULONG       StateSize;
    ULONG       VirtualizedSectors;
    ULONG64     InformationOffset[3];
    union {
        ULONG64 Mft2StartLcn;
        ULONG64 VirtualizedBlockOffset;
    };
} FVE_INFORMATION;

typedef struct _FVE_DATASET {
    ULONG       Size;
    ULONG       Version;
    ULONG       StartOffset;
    ULONG       EndOffset;
    CHAR        Identification[16];
    ULONG       NonceCounter;
    USHORT      FvekType;
    USHORT      _Unknown;
    ULONG64     CreationTime;
} FVE_DATASET;

typedef struct _FVE_DATUM {
    USHORT      Size;
    USHORT      Role;
    USHORT      Type;
    USHORT      Flags;
} FVE_DATUM;

typedef struct _FVE_VALIDATION {
    USHORT      Size;
    USHORT      Version;
    ULONG       Crc32;
    // FVE_DATUM   IntegrityCheck;
} FVE_VALIDATION;

/* ======== FVE datums ======== */

typedef struct _FVE_DATUM_SIMPLE {
    ULONG       Data;
} FVE_DATUM_SIMPLE;

typedef struct _FVE_DATUM_SIMPLE_LARGE {
    ULONG64     Data;
} FVE_DATUM_SIMPLE_LARGE;

typedef struct _FVE_DATUM_GUID {
    CHAR        Guid[16];
} FVE_DATUM_GUID;

typedef struct _FVE_DATUM_KEY {
    USHORT      KeyType;
    USHORT      KeyFlags;
    // CHAR        Data[];
} FVE_DATUM_KEY;

typedef struct _FVE_DATUM_UNICODE {
    // wchar       Text[];
} FVE_DATUM_UNICODE;

typedef struct _FVE_DATUM_STRETCH_KEY {
    USHORT      KeyType;
    USHORT      KeyFlags;
    CHAR        Salt[16];
} FVE_DATUM_STRETCH_KEY;

typedef struct _FVE_DATUM_USE_KEY {
    USHORT      KeyType;
    USHORT      KeyFlags;
} FVE_DATUM_USE_KEY;

typedef struct _FVE_NONCE {
  ULONG64       DateTime;
  ULONG         Counter;
} FVE_NONCE;

typedef struct _FVE_DATUM_AESCCM_ENC {
    FVE_NONCE   Nonce;
    CHAR        MAC[16];
    // CHAR        Data[];
} FVE_DATUM_AESCCM_ENC;

typedef struct _FVE_DATUM_TPM_ENC_BLOB {
    ULONG       PcrBitmap;
    // CHAR        Data[];
} FVE_DATUM_TPM_ENC_BLOB;

typedef struct _FVE_DATUM_VALIDATION_ENTRY {
    ULONG       _Unknown1;
    ULONG       _Unknown2;
    CHAR        Hash[32];
} FVE_DATUM_VALIDATION_ENTRY;

typedef struct _FVE_DATUM_VALIDATION_INFO {
    // FVE_DATUM_VALIDATION_ENTRY  AllowList[];
} FVE_DATUM_VALIDATION_INFO;

typedef struct _FVE_DATUM_VMK_INFO {
    CHAR        Identifier[16];
    ULONG64     DateTime;
    USHORT      _Unknown1;
    USHORT      Priority;
} FVE_DATUM_VMK_INFO;

typedef struct _FVE_DATUM_EXTERNAL_INFO {
    CHAR        Identifier[16];
    ULONG64     DateTime;
} FVE_DATUM_EXTERNAL_INFO;

typedef struct _FVE_DATUM_UPDATE {
    // Unknown
} FVE_DATUM_UPDATE;

typedef struct _FVE_DATUM_ERROR_LOG {
    // Unknown
} FVE_DATUM_ERROR_LOG;

typedef struct _FVE_DATUM_ASYM_ENC_BLOB {
    // CHAR        Data[];
} FVE_DATUM_ASYM_ENC_BLOB;

typedef struct _FVE_DATUM_EXPORTED_PUBLIC_KEY {
    // CHAR        Data[];
} FVE_DATUM_EXPORTED_PUBLIC_KEY;

typedef struct _FVE_DATUM_PUBLIC_KEY_INFO {
    // CHAR        Data[];
} FVE_DATUM_PUBLIC_KEY_INFO;

typedef struct _FVE_DATUM_VIRTUALIZATION_INFO {
    ULONG64     VirtualizedBlockOffset;
    ULONG64     VirtualizedBlockSize;
} FVE_DATUM_VIRTUALIZATION_INFO;

typedef struct _FVE_DATUM_CONCAT_HASH_KEY {
    // Unknown
} FVE_DATUM_CONCAT_HASH_KEY;

typedef struct _FVE_DATUM_BACKUP_INFO {
    // Unknown
} FVE_DATUM_BACKUP_INFO;

typedef struct _FVE_DATUM_AESCBC256_HMAC_SHA512_ENC {
    CHAR        Iv[16];
    CHAR        Mac[64];
    // CHAR        Data[];
} FVE_DATUM_AESCBC256_HMAC_SHA512_ENC;

/* ======== EOW structures ======== */

typedef struct _FVE_EOW_INFORMATION {
    CHAR        HeaderSignature[8];
    USHORT      HeaderSize;
    USHORT      Size;
    ULONG       SectorSize;
    ULONG       _Unknown1;
    ULONG       ChunkSize;
    ULONG       ConvLogSize;
    ULONG       _Unknown2;
    ULONG       RegionCount;
    ULONG       Crc32;
    ULONG64     EowOffset[2];
    ULONG64     BitmapOffsets[(Size - HeaderSize) / 8];
} FVE_EOW_INFORMATION;

typedef struct _FVE_EOW_BITMAP {
    CHAR        HeaderSignature[10];
    USHORT      HeaderSize;
    ULONG       Size;
    ULONG       _Unknown1;
    ULONG64     RegionOffset;
    ULONG64     RegionSize;
    ULONG64     ConvLogOffset;
    ULONG       RecordOffset[2];
    ULONG       RecordSize;
    ULONG       Crc32;
} FVE_EOW_BITMAP;

typedef struct _FVE_EOW_BITMAP_RECORD {
    CHAR        HeaderSignature[10];
    USHORT      HeaderSize;
    ULONG       Size;
    ULONG       BitmapSize;
    ULONG64     SequenceNumber;
    ULONG       Flags;
    ULONG       Crc32;
    // ULONG       Bitmap[];
} FVE_EOW_BITMAP_RECORD;
"""

c_bde = cstruct().load(bde_def)

FVE_STATE = c_bde.FVE_STATE
FVE_KEY_TYPE = c_bde.FVE_KEY_TYPE
FVE_KEY_FLAG = c_bde.FVE_KEY_FLAG
FVE_KEY_PROTECTOR = c_bde.FVE_KEY_PROTECTOR

FVE_DATUM_ROLE = c_bde.FVE_DATUM_ROLE
FVE_DATUM_TYPE = c_bde.FVE_DATUM_TYPE

# Volume signatures
BITLOCKER_SIGNATURE = b"-FVE-FS-"
BITLOCKER_TO_GO_SIGNATURE = b"MSWIN4.1"

EOW_SIGNATURE = b"FVE-EOW\x00"
EOW_BM_SIGNATURE = b"FVE-EOWBM\x00"
EOW_BR_SIGNATURE = b"FVE-EOWBR\x00"

CONV_MAGIC = b"FVEHDRLO"[::-1]

INFORMATION_OFFSET_GUID = UUID("4967d63b-2e29-4ad8-8399-f6a339e3d001")
EOW_INFORMATION_OFFSET_GUID = UUID("92a84d3b-dd80-4d0e-9e4e-b1e3284eaed8")

CIPHER_MAP = {
    FVE_KEY_TYPE.AES_128_DIFFUSER: "aes-cbc-128-elephant",
    FVE_KEY_TYPE.AES_256_DIFFUSER: "aes-cbc-256-elephant",
    FVE_KEY_TYPE.AES_128: "aes-cbc-128-eboiv",
    FVE_KEY_TYPE.AES_256: "aes-cbc-256-eboiv",
    FVE_KEY_TYPE.AES_XTS_128: "aes-xts-128-plain64",
    FVE_KEY_TYPE.AES_XTS_256: "aes-xts-256-plain64",
}



from __future__ import annotations

import hashlib
from binascii import crc32
from functools import cached_property
from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from Crypto.Cipher import AES
from dissect.util import ts

from dissect.fve.bde.c_bde import (
    BITLOCKER_SIGNATURE,
    FVE_DATUM_ROLE,
    FVE_DATUM_TYPE,
    FVE_KEY_FLAG,
    FVE_KEY_PROTECTOR,
    FVE_KEY_TYPE,
    FVE_STATE,
    c_bde,
)
from dissect.fve.exception import InvalidHeaderError

if TYPE_CHECKING:
    import datetime
    from collections.abc import Iterator


class Information:
    """Bitlocker Information.

    Parses Bitlocker Information and Dataset at a specified offset.

    Bitlocker Information consists of a small header, a Dataset and at least a CRC32 validation check.
    The CRC32 Validation information is positioned after the Information buffer.

    The ``StateOffset`` field contains the offset to a conversion log, but it also doubles as a "watermark",
    containing the offset up until where the Bitlocker encryption is active.
    The conversion log as pointed to by the ``StateOffset`` seems to only be used by older Bitlocker
    implementations. It looks like more modern implementations (Windows 10+) seem to prefer EOW.
    """

    def __init__(self, fh: BinaryIO, offset: int):
        self.offset = offset
        fh.seek(offset)

        self.header = c_bde.FVE_INFORMATION(fh)
        if self.header.Signature != BITLOCKER_SIGNATURE:
            raise InvalidHeaderError("Invalid BDE information signature")

        # Datums are lazily parsed so we can safely parse the dataset header
        self.dataset = Dataset(fh)

        fh.seek(offset)
        self._buf = fh.read(self.size)

        self.validation = Validation(fh)
        self._valid_checksum = crc32(self._buf) == self.validation.crc32

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"offset=0x{self.offset:x} current_state={self.current_state} next_state={self.next_state}>"
        )

    def is_valid(self) -> bool:
        """Validate the integrity of this Information block."""
        # TODO add sha256 check
        return self._valid_checksum

    def check_integrity(self, key: KeyDatum | bytes) -> bool:
        """Check the integrity of this Information block."""
        if self.validation.integrity_check:
            datum = self.validation.integrity_check.unbox(key)
            return hashlib.sha256(self._buf).digest() == datum.data
        return self.is_valid()

    @property
    def size(self) -> int:
        stored_size = self.header.HeaderSize
        if self.version >= 2:
            stored_size <<= 4
        return stored_size

    @property
    def version(self) -> int:
        return self.header.Version

    @property
    def current_state(self) -> FVE_STATE:
        return FVE_STATE(self.header.CurrentState)

    @property
    def next_state(self) -> FVE_STATE:
        return FVE_STATE(self.header.NextState)

    @property
    def state_offset(self) -> int:
        return self.header.StateOffset

    @property
    def state_size(self) -> int:
        return self.header.StateSize

    @property
    def virtualized_sectors(self) -> int:
        return self.header.VirtualizedSectors

    @property
    def virtualized_block_offset(self) -> int:
        return self.header.VirtualizedBlockOffset

    @property
    def information_offset(self) -> list[int]:
        return self.header.InformationOffset


class Validation:
    """Bitlocker Information Validation.

    The Validation structure is a small piece of data positioned after the Information buffer.
    It contains a CRC32 value of the entire Information buffer. It also contains an integrity check
    datum, which is a AES-CCM encrypted datum, encrypted with the same key that decrypts the FVEK.
    Decrypting the integrity check yields you a SHA256 digest, which must match the entire Information buffer.
    """

    def __init__(self, fh: BinaryIO):
        self.validation = c_bde.FVE_VALIDATION(fh)
        self.integrity_check = None
        if self.version >= 2:  # I think
            self.integrity_check = Datum.from_fh(fh)

    @property
    def version(self) -> int:
        return self.validation.Version

    @property
    def crc32(self) -> int:
        return self.validation.Crc32


class Dataset:
    """Bitlocker Information Dataset.

    The dataset is a simple data structure, consisting of a small header and one or more "datum".
    Each datum has a role and type, and you can query the dataset for datums with a specific role or type.
    Querying the dataset means iterating the datum array until you found the datum you're looking for.
    """

    def __init__(self, fh: BinaryIO):
        offset = fh.tell()
        self.header = c_bde.FVE_DATASET(fh)
        self.identifier = UUID(bytes_le=self.header.Identification)

        fh.seek(offset)
        self._buf = fh.read(self.header.Size)

    def __iter__(self) -> Iterator[Datum]:
        yield from self.data

    @cached_property
    def data(self) -> list[Datum]:
        """Return the list of Datum in this Dataset."""
        result = []

        buf = BytesIO(memoryview(self._buf)[self.header.StartOffset :])
        remaining = self.header.EndOffset - self.header.StartOffset
        while remaining >= Datum.MINIMAL_SIZE:
            datum = Datum.from_fh(buf)
            result.append(datum)

            remaining -= datum.size

        return result

    @property
    def fvek_type(self) -> FVE_KEY_TYPE:
        return FVE_KEY_TYPE(self.header.FvekType)

    def find_datum(self, role: FVE_DATUM_ROLE, type_: FVE_DATUM_TYPE) -> Iterator[Datum]:
        """Find one or more datum specified by role and type."""
        for datum in self:
            if (datum.role == role or role is None) and (datum.type == type_ or type_ is None):
                yield datum

    def find_description(self) -> str | None:
        """Find the description datum."""
        for datum in self.find_datum(FVE_DATUM_ROLE.DESCRIPTION, FVE_DATUM_TYPE.UNICODE):
            return datum.text
        return None

    def find_virtualization_info(self) -> VirtualizationInfoDatum | None:
        """Find the virtualization info datum."""
        for datum in self.find_datum(FVE_DATUM_ROLE.VIRTUALIZATION_INFO, FVE_DATUM_TYPE.VIRTUALIZATION_INFO):
            return datum
        return None

    def find_startup_key(self) -> ExternalInfoDatum | None:
        """Find the external startup/recovery key information."""
        for datum in self.find_datum(FVE_DATUM_ROLE.STARTUP_KEY, FVE_DATUM_TYPE.EXTERNAL_INFO):
            return datum
        return None

    def find_fvek(self) -> AesCcmEncryptedDatum | None:
        """Find the encrypted FVEK."""
        for datum in self.find_datum(FVE_DATUM_ROLE.FULL_VOLUME_ENCRYPTION_KEY, FVE_DATUM_TYPE.AES_CCM_ENCRYPTED_KEY):
            return datum
        return None

    def find_vmk(
        self,
        protector_type: FVE_KEY_PROTECTOR | None = None,
        min_priority: int = 0x0000,
        max_priority: int = 0x7FFF,
        mask: int = 0xFF00,
    ) -> Iterator[VmkInfoDatum]:
        """Find one or more VMK datum specified by key priority."""
        for datum in self.find_datum(FVE_DATUM_ROLE.VOLUME_MASTER_KEY_INFO, FVE_DATUM_TYPE.VOLUME_MASTER_KEY_INFO):
            if datum.priority.value < min_priority or datum.priority.value > max_priority:
                continue

            if protector_type is None or datum.priority & mask == protector_type:
                yield datum

    def find_clear_vmk(self) -> VmkInfoDatum | None:
        """Find the clear key VMK (for paused volumes)."""
        for vmk in self.find_vmk(FVE_KEY_PROTECTOR.CLEAR, max_priority=0xFF, mask=0x0000):
            return vmk
        return None

    def find_external_vmk(self) -> Iterator[VmkInfoDatum]:
        """Find the external VMK."""
        yield from self.find_vmk(FVE_KEY_PROTECTOR.EXTERNAL)

    def find_recovery_vmk(self) -> Iterator[VmkInfoDatum]:
        """Find the recovery VMK."""
        yield from self.find_vmk(FVE_KEY_PROTECTOR.RECOVERY_PASSWORD)

    def find_passphrase_vmk(self) -> Iterator[VmkInfoDatum]:
        """Find the passphrase VMK."""
        yield from self.find_vmk(FVE_KEY_PROTECTOR.PASSPHRASE)


class Datum:
    """Bitlocker Dataset Datum.

    A Datum is the main metadata structure in Bitlocker. It's a small data structure, specifying a
    size, role and type, followed by the necessary data to interpret that datum type.

    Datums can be "complex", in which case they can contain nested datums. These nested datums always
    have the PROPERTY role.

    Datums can also have a data segment. A data segment is present if a datum is not complex, but contains
    data beyond the size of that datums' type structure.

    Originally, this information is stored in a table, also containing a type's minimal size. This implementation
    doesn't currently do that, instead relying on the reading from a file handle with cstruct. Whatever is left
    on the file handle is the data segment.
    """

    __struct__ = None
    __complex__ = False

    MINIMAL_SIZE = len(c_bde.FVE_DATUM)

    def __init__(self, fh: BinaryIO):
        self.header = c_bde.FVE_DATUM(fh)
        self._data = fh.read(self.data_size)

        buf = BytesIO(self._data)
        self._datum = self.__struct__(buf) if self.__struct__ else None
        self.data_segment = buf.read() if not self.__complex__ else None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} type={self.type.name}>"

    @property
    def role(self) -> FVE_DATUM_ROLE:
        return FVE_DATUM_ROLE(self.header.Role)

    @property
    def type(self) -> FVE_DATUM_TYPE:
        return FVE_DATUM_TYPE(self.header.Type)

    @property
    def size(self) -> int:
        return self.header.Size

    @property
    def data_size(self) -> int:
        return self.size - self.MINIMAL_SIZE

    @cached_property
    def properties(self) -> list[Datum]:
        """Return a list of property datum if this datum is complex."""
        result = []

        if self.__complex__:
            remaining = self.data_size - len(self.__struct__)
            buf = BytesIO(memoryview(self._data)[len(self.__struct__) :])
            while remaining >= self.MINIMAL_SIZE:
                nested = Datum.from_fh(buf)
                result.append(nested)

                remaining -= nested.size

        return result

    @classmethod
    def from_fh(cls, fh: BinaryIO) -> Datum:
        """Read a datum from a file handle."""
        offset = fh.tell()
        header = c_bde.FVE_DATUM(fh)
        fh.seek(offset)

        return DATUM_TYPE_MAP.get(FVE_DATUM_TYPE(header.Type), Datum)(fh)

    @classmethod
    def from_bytes(cls, buf: bytes) -> Datum:
        """Read a datum from raw bytes."""
        return cls.from_fh(BytesIO(buf))

    def find_property(self, type_: FVE_DATUM_TYPE | None) -> Iterator[Datum]:
        """Find one or more datum with a specified type within the properties."""
        for datum in self.properties:
            if datum.type == type_ or type_ is None:
                yield datum


class SimpleDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_SIMPLE

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} data={self.data}>"

    @property
    def data(self) -> int:
        return self._datum.Data


class SimpleLargeDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_SIMPLE_LARGE

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} data={self.data}>"

    @property
    def data(self) -> int:
        return self._datum.Data


class GuidDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_GUID

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} guid={self.guid}>"

    @property
    def guid(self) -> UUID:
        return UUID(bytes_le=self._datum.Guid)


class KeyDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_KEY

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} key_type={self.key_type} key_flags={self.key_flags}>"

    @property
    def key_type(self) -> FVE_KEY_TYPE:
        return FVE_KEY_TYPE(self._datum.KeyType)

    @property
    def key_flags(self) -> FVE_KEY_FLAG:
        return FVE_KEY_FLAG(self._datum.KeyFlags)

    @property
    def data(self) -> bytes:
        return self._data[len(KeyDatum.__struct__) :]


class UnicodeDatum(Datum):
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} text={self.text}>"

    @property
    def text(self) -> str:
        return self._data.decode("utf-16-le").rstrip("\x00")


class StretchKeyDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_STRETCH_KEY
    __complex__ = True

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} key_type={self.key_type} key_flags={self.key_flags}>"

    @property
    def key_type(self) -> FVE_KEY_TYPE:
        return FVE_KEY_TYPE(self._datum.KeyType)

    @property
    def key_flags(self) -> FVE_KEY_FLAG:
        return FVE_KEY_FLAG(self._datum.KeyFlags)

    @property
    def salt(self) -> bytes:
        return self._datum.Salt


class UseKeyDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_USE_KEY
    __complex__ = True

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} key_type={self.key_type} key_flags={self.key_flags}>"

    @property
    def key_type(self) -> FVE_KEY_TYPE:
        return FVE_KEY_TYPE(self._datum.KeyType)

    @property
    def key_flags(self) -> FVE_KEY_FLAG:
        return FVE_KEY_FLAG(self._datum.KeyFlags)


class AesCcmEncryptedDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_AESCCM_ENC

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} role={self.role.name} "
            f"nonce_time={self.nonce_time} nonce_counter={self.nonce_counter}>"
        )

    @property
    def nonce(self) -> bytes:
        return self._data[: len(c_bde.FVE_NONCE)]

    @property
    def nonce_time(self) -> datetime.datetime | int:
        try:
            return ts.wintimestamp(self._datum.Nonce.DateTime)
        except ValueError:
            return self._datum.Nonce.DateTime

    @property
    def nonce_counter(self) -> int:
        return self._datum.Nonce.Counter

    @property
    def mac(self) -> bytes:
        return self._datum.MAC

    @property
    def data(self) -> bytes:
        return self._data[len(self.__struct__) :]

    def unbox(self, key: KeyDatum | bytes) -> Datum:
        key = key.data if isinstance(key, KeyDatum) else key
        cipher = AES.new(key, AES.MODE_CCM, nonce=self.nonce)
        decrypted_data = cipher.decrypt_and_verify(self.data, self.mac)
        return Datum.from_bytes(decrypted_data)


class TpmEncryptedBlobDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_TPM_ENC_BLOB

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} pcr_bitmap={self.pcr_bitmap}>"

    @property
    def pcr_bitmap(self) -> int:
        return self._datum.PcrBitmap

    @property
    def data(self) -> bytes:
        return self._data[len(self.__struct__) :]


class ValidationEntry:
    def __init__(self, fh: BinaryIO):
        self._entry = c_bde.FVE_DATUM_VALIDATION_ENTRY(fh)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} hash={self.hash}>"

    @property
    def hash(self) -> bytes:
        return self._entry.Hash


class ValidationInfoDatum(Datum):
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name} allow_list={self.allow_list}>"

    @property
    def allow_list(self) -> list[ValidationEntry]:
        fh = BytesIO(self._data)
        return [ValidationEntry(fh) for _ in range(len(self._data) // len(c_bde.FVE_DATUM_VALIDATION_ENTRY))]


class VmkInfoDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_VMK_INFO
    __complex__ = True

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} role={self.role.name} identifier={self.identifier} "
            f"datetime={self.datetime} priority={self.priority}>"
        )

    @property
    def identifier(self) -> UUID:
        return UUID(bytes_le=self._datum.Identifier)

    @property
    def datetime(self) -> datetime.datetime:
        return ts.wintimestamp(self._datum.DateTime)

    @property
    def priority(self) -> FVE_KEY_PROTECTOR:
        return FVE_KEY_PROTECTOR(self._datum.Priority)

    def decrypt(self, key: KeyDatum | bytes) -> KeyDatum:
        encrypted_key = self.aes_ccm_encrypted_key()
        return encrypted_key.unbox(key)

    def label(self) -> str | None:
        for datum in self.find_property(FVE_DATUM_TYPE.UNICODE):
            return datum.text
        return None

    def asymmetric_encrypted_key(self) -> AsymmetricEncryptedDatum | None:
        for datum in self.find_property(FVE_DATUM_TYPE.ASYMMETRIC_ENCRYPTED_KEY):
            return datum
        return None

    def exported_key(self) -> ExportedPublicKeyDatum | None:
        for datum in self.find_property(FVE_DATUM_TYPE.EXPORTED_KEY):
            return datum
        return None

    def tpm_encrypted_blob(self) -> TpmEncryptedBlobDatum | None:
        for datum in self.find_property(FVE_DATUM_TYPE.TPM_ENCRYPTED_BLOB):
            return datum
        return None

    def aes_ccm_encrypted_key(self) -> AesCcmEncryptedDatum | None:
        for datum in self.find_property(FVE_DATUM_TYPE.AES_CCM_ENCRYPTED_KEY):
            return datum
        return None

    def public_key_info(self) -> PublicKeyInfoDatum | None:
        for datum in self.find_property(FVE_DATUM_TYPE.PUBLIC_KEY_INFO):
            return datum
        return None

    def use_keys(self) -> list[UseKeyDatum]:
        return list(self.find_property(FVE_DATUM_TYPE.USE_KEY))

    def use_key(self, key_type: FVE_KEY_TYPE) -> UseKeyDatum | None:
        for datum in self.use_keys():
            if key_type is None or datum.key_type == key_type:
                return datum
        return None

    def stretch_keys(self) -> list[StretchKeyDatum]:
        return list(self.find_property(FVE_DATUM_TYPE.STRETCH_KEY))

    def stretch_key(self, key_type: FVE_KEY_TYPE) -> StretchKeyDatum | None:
        for datum in self.stretch_keys():
            if key_type is None or datum.key_type == key_type:
                return datum
        return None

    def clear_key(self) -> KeyDatum | None:
        for datum in self.find_property(FVE_DATUM_TYPE.KEY):
            return datum
        return None

    def is_enhanced_pin(self) -> bool:
        for stretch_key in self.stretch_keys():
            if stretch_key.key_type == FVE_KEY_TYPE.AES_CCM_256_2 and stretch_key.key_flags & FVE_KEY_FLAG.ENHANCED_PIN:
                return True
        return False

    def is_enhanced_crypto(self) -> bool:
        for stretch_key in self.stretch_keys():
            if (
                stretch_key.key_type == FVE_KEY_TYPE.AES_CCM_256_2
                and stretch_key.key_flags & FVE_KEY_FLAG.ENHANCED_CRYPTO
            ):
                return True
        return False

    def uses_pbkdf2(self) -> bool:
        for stretch_key in self.stretch_keys():
            if (
                stretch_key.type in (FVE_KEY_TYPE.STRETCH_KEY, FVE_KEY_TYPE.STRETCH_KEY_1, FVE_KEY_TYPE.AES_CCM_256_2)
                and stretch_key.key_flags & FVE_KEY_FLAG.PBKDF2
            ):
                return True
        return False


class ExternalInfoDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_EXTERNAL_INFO
    __complex__ = True

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} role={self.role.name} identifier={self.identifier} datetime={self.datetime}>"
        )

    @property
    def identifier(self) -> UUID:
        return UUID(bytes_le=self._datum.Identifier)

    @property
    def datetime(self) -> datetime.datetime:
        return ts.wintimestamp(self._datum.DateTime)

    def label(self) -> str | None:
        for datum in self.find_property(FVE_DATUM_TYPE.UNICODE):
            return datum.text
        return None

    def external_key(self) -> KeyDatum | None:
        for datum in self.find_property(FVE_DATUM_TYPE.KEY):
            return datum
        return None


class UpdateDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_UPDATE
    __complex__ = True

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name}>"


class ErrorLogDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_ERROR_LOG

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name}>"


class AsymmetricEncryptedDatum(Datum):
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name}>"

    @property
    def data(self) -> bytes:
        return self._data


class ExportedPublicKeyDatum(Datum):
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name}>"

    @property
    def data(self) -> bytes:
        return self._data


class PublicKeyInfoDatum(Datum):
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name}>"

    @property
    def data(self) -> bytes:
        return self._data


class VirtualizationInfoDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_VIRTUALIZATION_INFO

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} role={self.role.name} "
            f"virtualized_block_offset=0x{self.virtualized_block_offset:x} "
            f"virtualized_block_size=0x{self.virtualized_block_size:x}>"
        )

    @property
    def virtualized_block_offset(self) -> int:
        return self._datum.VirtualizedBlockOffset

    @property
    def virtualized_block_size(self) -> int:
        return self._datum.VirtualizedBlockSize


class ConcatHashKeyDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_CONCAT_HASH_KEY

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name}>"


class BackupInfoDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_BACKUP_INFO

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name}>"


class AesCbc256HmacSha512EncryptedDatum(Datum):
    __struct__ = c_bde.FVE_DATUM_AESCBC256_HMAC_SHA512_ENC

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} role={self.role.name}>"

    @property
    def iv(self) -> bytes:
        return self._datum.Iv

    @property
    def mac(self) -> bytes:
        return self._datum.Mac

    @property
    def data(self) -> bytes:
        return self._data[len(self.__struct__) :]


DATUM_TYPE_MAP = {
    FVE_DATUM_TYPE.KEY: KeyDatum,
    FVE_DATUM_TYPE.UNICODE: UnicodeDatum,
    FVE_DATUM_TYPE.STRETCH_KEY: StretchKeyDatum,
    FVE_DATUM_TYPE.USE_KEY: UseKeyDatum,
    FVE_DATUM_TYPE.AES_CCM_ENCRYPTED_KEY: AesCcmEncryptedDatum,
    FVE_DATUM_TYPE.TPM_ENCRYPTED_BLOB: TpmEncryptedBlobDatum,
    FVE_DATUM_TYPE.VALIDATION_INFO: ValidationInfoDatum,
    FVE_DATUM_TYPE.VOLUME_MASTER_KEY_INFO: VmkInfoDatum,
    FVE_DATUM_TYPE.EXTERNAL_INFO: ExternalInfoDatum,
    FVE_DATUM_TYPE.UPDATE: UpdateDatum,
    FVE_DATUM_TYPE.ERROR_LOG: ErrorLogDatum,
    FVE_DATUM_TYPE.ASYMMETRIC_ENCRYPTED_KEY: AsymmetricEncryptedDatum,
    FVE_DATUM_TYPE.EXPORTED_KEY: ExportedPublicKeyDatum,
    FVE_DATUM_TYPE.PUBLIC_KEY_INFO: PublicKeyInfoDatum,
    FVE_DATUM_TYPE.VIRTUALIZATION_INFO: VirtualizationInfoDatum,
    FVE_DATUM_TYPE.SIMPLE_1: SimpleDatum,
    FVE_DATUM_TYPE.SIMPLE_2: SimpleDatum,
    FVE_DATUM_TYPE.CONCAT_HASH_KEY: ConcatHashKeyDatum,
    FVE_DATUM_TYPE.SIMPLE_3: SimpleDatum,
    FVE_DATUM_TYPE.SIMPLE_LARGE: SimpleLargeDatum,
    FVE_DATUM_TYPE.BACKUP_INFO: BackupInfoDatum,
}



# References:
# - fvevol.sys

from __future__ import annotations

from binascii import crc32
from functools import cached_property
from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

from dissect.fve.bde.c_bde import (
    EOW_BM_SIGNATURE,
    EOW_BR_SIGNATURE,
    EOW_SIGNATURE,
    c_bde,
)
from dissect.fve.exception import InvalidHeaderError

if TYPE_CHECKING:
    from collections.abc import Iterator


class EowInformation:
    """Bitlocker EOW Information."""

    def __init__(self, fh: BinaryIO, offset: int):
        self.fh = fh
        self.offset = offset
        fh.seek(offset)

        self.header = c_bde.FVE_EOW_INFORMATION(fh)
        if self.header.HeaderSignature != EOW_SIGNATURE:
            raise InvalidHeaderError("Invalid EOW information signature")

        _crc32 = self.header.Crc32
        self.header.Crc32 = 0
        self._valid_checksum = crc32(self.header.dumps()) == _crc32
        self.header.Crc32 = _crc32

    def is_valid(self) -> bool:
        return self._valid_checksum

    @property
    def size(self) -> int:
        return self.header.Size

    @property
    def chunk_size(self) -> int:
        return self.header.ChunkSize

    @property
    def conv_log_size(self) -> int:
        return self.header.ConvLogSize

    @cached_property
    def bitmaps(self) -> list[EowBitmap]:
        return [EowBitmap(self.fh, offset) for offset in self.header.BitmapOffsets]


class EowBitmap:
    """Bitlocker EOW Bitmap.

    A bitmap contains multiple bitmap records, but only one record is active. The active record is
    determined by the Lsn field in the header. The record with the highest Lsn is the active record.

    It looks like the number of bitmap records is hardcoded to 2, but let's keep the implementation
    flexible.
    """

    def __init__(self, fh: BinaryIO, offset: int):
        self.fh = fh
        self.offset = offset
        fh.seek(offset)

        self.header = c_bde.FVE_EOW_BITMAP(fh)
        if self.header.HeaderSignature != EOW_BM_SIGNATURE:
            raise ValueError("Invalid EOW bitmap signature")

        _crc32 = self.header.Crc32
        self.header.Crc32 = 0
        remainder = fh.read(self.header.RecordOffset[0] - self.header.HeaderSize)
        self._valid_checksum = crc32(self.header.dumps() + remainder) == _crc32
        self.header.Crc32 = _crc32

        self._record_data = fh.read(self.header.Size - self.header.RecordOffset[0])

    def __repr__(self) -> str:
        return f"<EowBitmap region_offset=0x{self.region_offset:x} region_size=0x{self.region_size:x}>"

    def is_valid(self) -> bool:
        return self._valid_checksum

    def runs(self, chunk: int, length: int) -> Iterator[tuple[int, int]]:
        yield from self.active_record.runs(chunk, length)

    @property
    def size(self) -> int:
        return self.header.Size

    @property
    def region_offset(self) -> int:
        return self.header.RegionOffset

    @property
    def region_size(self) -> int:
        return self.header.RegionSize

    @property
    def conv_log_offset(self) -> int:
        return self.header.ConvLogOffset

    @cached_property
    def active_record(self) -> EowBitmapRecord:
        latest_record = None

        for record in self.records:
            if latest_record is None:
                latest_record = record
                continue

            if record.sequence_number > latest_record.sequence_number:
                latest_record = record

        return latest_record

    @cached_property
    def records(self) -> list[EowBitmapRecord]:
        result = []

        buf = BytesIO(self._record_data)
        base = self.header.RecordOffset[0]
        for offset in self.header.RecordOffset:
            buf.seek(offset - base)
            result.append(EowBitmapRecord(buf))

        return result


class EowBitmapRecord:
    """Bitlocker EOW Bitmap Record.

    The record holding the actual bitmap. Each bit indicates a chunk with the size defined by
    the EOW information. The Lsn is the sequence number of that record.

    The flags are currently unknown, but seem related to an encrypted/decrypted state.
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.header = c_bde.FVE_EOW_BITMAP_RECORD(fh)
        if self.header.HeaderSignature != EOW_BR_SIGNATURE:
            raise ValueError("Invalid EOW bitmap record signature")

        _crc32 = self.header.Crc32
        self.header.Crc32 = 0
        self._data = memoryview(fh.read(self.header.Size - self.header.HeaderSize))
        self._valid_checksum = crc32(self.header.dumps() + self._data) == _crc32
        self.header.Crc32 = _crc32

    def __repr__(self) -> str:
        return f"<EowBitmapRecord sequence_number={self.sequence_number} bitmap_size=0x{self.bitmap_size:x}>"

    def is_valid(self) -> bool:
        return self._valid_checksum

    def runs(self, chunk: int, length: int) -> Iterator[tuple[int, int]]:
        yield from _iter_bitmap(self.bitmap, self.bitmap_size, chunk, length)

    @property
    def size(self) -> int:
        return self.header.Size

    @property
    def bitmap(self) -> bytes:
        return self._data

    @property
    def bitmap_size(self) -> int:
        return self.header.BitmapSize

    @property
    def sequence_number(self) -> int:
        return self.header.SequenceNumber


def _iter_bitmap(bitmap: bytes, size: int, start: int, count: int) -> Iterator[tuple[int, int]]:
    byte_idx, bit_idx = divmod(start, 8)
    remaining_bits = size - start
    current_bit = (bitmap[byte_idx] & (1 << bit_idx)) >> bit_idx
    current_count = 0

    for byte in bitmap[byte_idx:]:
        if count == 0 or remaining_bits == 0:
            break

        if (current_bit, byte) == (0, 0) or (current_bit, byte) == (1, 0xFF):
            max_count = min(count, remaining_bits, 8 - bit_idx)
            current_count += max_count
            remaining_bits -= max_count
            count -= max_count
            bit_idx = 0
        else:
            for cur_bit_idx in range(bit_idx, min(count, remaining_bits, 8)):
                bit_set = (byte & (1 << cur_bit_idx)) >> cur_bit_idx

                if bit_set == current_bit:
                    current_count += 1
                else:
                    yield (current_bit, current_count)
                    current_bit = bit_set
                    current_count = 1

                remaining_bits -= 1
                count -= 1

    if current_count:
        yield (current_bit, current_count)



from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from dissect.fve.crypto import argon2, create_cipher
from dissect.fve.crypto.dmcrypt import CryptStream
from dissect.fve.luks import af
from dissect.fve.luks.c_luks import (
    LUKS2_MAGIC_1ST,
    LUKS2_MAGIC_2ND,
    SECONDARY_HEADER_OFFSETS,
    c_luks,
)
from dissect.fve.luks.metadata import Metadata

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.fve.luks.metadata import Digest, Keyslot, Segment


class LUKS:
    """LUKS disk encryption.

    References:
        - https://gitlab.com/cryptsetup/cryptsetup
        - https://gitlab.com/cryptsetup/cryptsetup/-/blob/main/docs/on-disk-format-luks2.pdf
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.header = None
        self.header1 = None
        self.header2 = None

        first_offset, second_offset, version = find_luks_headers(fh)
        if version is None:
            raise ValueError("Not a LUKS volume")

        if version == 1:
            header_struct = c_luks.luks_phdr
        elif version == 2:
            header_struct = c_luks.luks2_hdr_disk
        else:
            raise ValueError(f"Unsupported LUKS version: {version}")

        if first_offset is not None:
            fh.seek(first_offset)
            self.header1 = header_struct(fh)
            self.header = self.header1

        if second_offset is not None:
            fh.seek(second_offset)
            self.header2 = header_struct(fh)

        self.header = self.header2 or self.header1

        # LUKS1
        self.cipher_name = None
        self.cipher_mode = None
        self.hash_spec = None

        # LUKS2
        self.label = None
        self.checksum_algorithm = None
        self.metadata, self.metadata1, self.metadata2 = None, None, None

        self.uuid = UUID(self.header.uuid.strip(b"\x00").decode())

        if self.header.version == 1:
            # LUKS1
            self.metadata = Metadata.from_luks1_header(self.header)
        else:
            # LUKS2
            self.label = self.header.label.strip(b"\x00").decode()
            self.checksum_algorithm = self.header.checksum_alg.strip(b"\x00").decode()

            self.metadata1 = None
            if self.header is self.header1:
                json_area1 = fh.read(self.header1.hdr_size - 4096).rstrip(b"\x00").decode()
                self.metadata1 = Metadata.from_json(json_area1)
                self.header2 = c_luks.luks2_hdr_disk(fh)

            json_area2 = fh.read(self.header2.hdr_size - 4096).rstrip(b"\x00").decode()
            self.metadata2 = Metadata.from_json(json_area2)

            self.metadata = self.metadata1 or self.metadata2

        self._active_volume_key = None
        self._active_keyslot_id = None

    @property
    def unlocked(self) -> bool:
        return self._active_volume_key is not None

    @property
    def keyslots(self) -> dict[int, Keyslot]:
        return self.metadata.keyslots

    def unlock(self, key: bytes, keyslot: int) -> None:
        """Unlock the volume with the volume encryption key."""
        if not self._verify_volume_key(key, keyslot):
            raise ValueError(f"Invalid volume key for keyslot {keyslot}")
        self._active_volume_key = key
        self._active_keyslot_id = keyslot

    def unlock_with_key_file(self, path: Path, offset: int = 0, size: int = -1, keyslot: int | None = None) -> None:
        with path.open("rb") as fh:
            self.unlock_with_key_fh(fh, offset, size, keyslot)

    def unlock_with_key_fh(self, fh: BinaryIO, offset: int = 0, size: int = -1, keyslot: int | None = None) -> None:
        fh.seek(offset)
        self._unlock_passphrase(fh.read(size), keyslot)

    def unlock_with_passphrase(self, passphrase: str | bytes, keyslot: int | None = None) -> None:
        """Unlock this volume with a passphrase and optional keyslot hint."""
        if isinstance(passphrase, str):
            self._unlock_passphrase(passphrase.encode(), keyslot)
        else:
            self._unlock_passphrase(passphrase, keyslot)

    def _unlock_passphrase(self, passphrase: bytes, keyslot: int | None = None) -> None:
        """Unlock this volume with a passphrase and optional keyslot hint."""
        keyslots = (
            [(keyslot, self.metadata.keyslots[keyslot])] if keyslot is not None else self.metadata.keyslots.items()
        )

        idx = None
        vk = None
        errors = []
        for idx, keyslot in keyslots:
            try:
                key = derive_passphrase_key(passphrase, keyslot)
            except Exception as exc:
                errors.append((idx, exc))
                continue

            try:
                vk = self._unlock_volume_key(key, idx)
            except Exception as exc:
                errors.append((idx, exc))
                continue

            try:
                self.unlock(vk, idx)
                break
            except ValueError:
                continue
        else:
            if errors:
                msg = "\n".join(f"{idx}: {exc}" for idx, exc in errors)
                raise ValueError(f"No valid keyslot found, but there were errors for the following keyslots:\n{msg}")
            raise ValueError("No valid keyslot found")

    def _unlock_volume_key(self, key: bytes, keyslot: int) -> None:
        """Unlock the volume key using the given encryption key and keyslot."""
        keyslot_obj = self.metadata.keyslots[keyslot]

        self.fh.seek(keyslot_obj.area.offset)
        area = self.fh.read(keyslot_obj.key_size * keyslot_obj.af.stripes)

        cipher = create_cipher(keyslot_obj.area.encryption, key, keyslot_obj.area.key_size * 8)
        return af.merge(
            cipher.decrypt(area),
            keyslot_obj.key_size,
            keyslot_obj.af.stripes,
            keyslot_obj.af.hash,
        )

    def _verify_volume_key(self, key: bytes, keyslot: int) -> None:
        """Verify the given key for the given keyslot."""
        digest = self.find_digest(keyslot)
        if digest.type == "pbkdf2":
            result = hashlib.pbkdf2_hmac(digest.hash, key, digest.salt, digest.iterations, len(digest.digest))
        else:
            # Only the pbkdf2 type is supported in LUKS2
            raise NotImplementedError(f"Unsupported digest algorithm: {digest.type}")

        return result == digest.digest

    def find_digest(self, keyslot: int) -> Digest:
        """Find digest metadata corresponding to the given keyslot."""
        digests = [digest for digest in self.metadata.digests.values() if keyslot in digest.keyslots]
        if not digests:
            raise ValueError(f"No digest found for keyslot {keyslot}")

        return digests[0]

    def find_segment(self, keyslot: int) -> Segment:
        """Find segment metadata corresponding to the given keyslot."""
        digest = self.find_digest(keyslot)
        segments = [segment for segment_id, segment in self.metadata.segments.items() if segment_id in digest.segments]
        if not segments:
            raise ValueError(f"No segment found for keyslot {keyslot}")

        if len(segments) > 1:
            raise NotImplementedError(f"Keyslot {keyslot} has more than one segment")

        return segments[0]

    def open(self) -> CryptStream:
        """Open this volume and return a readable (decrypted) stream."""
        if not self.unlocked:
            raise ValueError("Volume is locked")

        # Technically LUKS supports multiple segments, but practically it only ever has one
        # Don't bother with supporting multiple segments for now
        segment = self.find_segment(self._active_keyslot_id)

        return CryptStream(
            self.fh,
            segment.encryption,
            self._active_volume_key,
            self.metadata.keyslots[self._active_keyslot_id].key_size * 8,
            segment.offset,
            segment.size,
            segment.iv_tweak,
            segment.sector_size,
        )


def derive_passphrase_key(passphrase: bytes, keyslot: Keyslot) -> bytes:
    """Derive a key from a passphrase with the given keyslot KDF information.

    Args:
        passphrase: The passphrase to derive a key from.
        keyslot: The keyslot to use for the derivation.
    """
    kdf = keyslot.kdf

    if kdf.type == "pbkdf2":
        return hashlib.pbkdf2_hmac(kdf.hash, passphrase, kdf.salt, kdf.iterations, keyslot.key_size)

    if kdf.type.startswith("argon2"):
        return argon2.hash_secret_raw(passphrase, kdf.salt, kdf.time, kdf.memory, kdf.cpus, keyslot.key_size, kdf.type)

    raise NotImplementedError(f"Unsupported kdf algorithm: {kdf.type}")


def find_luks_headers(fh: BinaryIO) -> tuple[int | None, int | None, int | None]:
    stored_position = fh.tell()

    fh.seek(0)
    first_header = None
    second_header = None
    version = None

    if fh.read(c_luks.LUKS2_MAGIC_L) == LUKS2_MAGIC_1ST:
        first_header = 0
        version = int.from_bytes(fh.read(2), "big")

    for offset in SECONDARY_HEADER_OFFSETS:
        fh.seek(offset)
        if fh.read(c_luks.LUKS2_MAGIC_L) == LUKS2_MAGIC_2ND:
            second_header = offset
            version = int.from_bytes(fh.read(2), "big")
            break

    fh.seek(stored_position)
    return first_header, second_header, version


def is_luks_volume(fh: BinaryIO) -> bool:
    """Return whether the file-like object is a LUKS volume."""
    _, _, version = find_luks_headers(fh)
    return version is not None



from __future__ import annotations

import hashlib
import platform
import sys
from typing import TYPE_CHECKING, Any

from Crypto.Cipher import AES
from Crypto.Util import _raw_api

from dissect.fve.crypto import elephant
from dissect.fve.crypto.base import DECRYPT, ENCRYPT, IV, Cipher

if platform.python_implementation() == "CPython":
    # On CPython, our own "pure Python" XOR is somehow faster than the one from pycryptodome
    from dissect.fve.crypto.util import xor
else:
    # On PyPy the opposite is true, and also just use this as the default fallback
    from Crypto.Util.strxor import strxor as xor


if TYPE_CHECKING:
    from collections.abc import Callable

POINTER_SIZE = 8 if sys.maxsize > 2**32 else 4


if _raw_api.backend == "cffi":

    def get_iv_view(cipher: AES.CbcMode | AES.EcbMode, size: int) -> memoryview:
        return _raw_api.ffi.cast(_raw_api.uint8_t_type, cipher._state.get() + POINTER_SIZE)[0:size]

elif _raw_api.backend == "ctypes":
    import ctypes

    def get_iv_view(cipher: AES.CbcMode | AES.EcbMode, size: int) -> memoryview:
        return ctypes.cast(cipher._state.get().value + POINTER_SIZE, ctypes.POINTER(ctypes.c_char * size))[0]

else:

    def get_iv_view(cipher: AES.CbcMode | AES.EcbMode, size: int) -> memoryview:
        raise NotImplementedError("Unsupported pycryptodome backend")


# Sanity check if fast IV is available
def _fast_iv_works() -> bool:
    try:
        magic = b"\x69" * 16
        cipher = AES.new(b"\x00" * 16, AES.MODE_CBC, iv=magic)
        return _raw_api.get_raw_buffer(get_iv_view(cipher, 16)) == magic
    except Exception:
        return False


FAST_IV = _fast_iv_works()


class EcbMode(Cipher):
    """ECB mode implementation for FVE crypto."""

    def __init__(
        self,
        factory: Any,
        key: bytes,
        key_size: int,
        iv_mode: type[IV],
        iv_options: str,
        sector_size: int = 512,
        iv_sector_size: int = 512,
    ):
        if key_size not in (128, 256):
            raise ValueError(f"Incorrect key size for ECB mode ({key_size} bits)")
        super().__init__(key, key_size, factory.block_size, iv_mode, iv_options, sector_size, iv_sector_size)

        self._cipher = AES.new(key[: self.key_size_bytes], AES.MODE_ECB)

    def _crypt_sector(self, mode: int, buffer: bytearray, iv: bytes) -> None:
        (self._cipher.encrypt if mode == ENCRYPT else self._cipher.decrypt)(buffer, output=buffer)


class CbcMode(Cipher):
    """CBC mode implementation for FVE crypto."""

    def __init__(
        self,
        factory: Any,
        key: bytes,
        key_size: int,
        iv_mode: type[IV],
        iv_options: str,
        sector_size: int = 512,
        iv_sector_size: int = 512,
    ):
        if key_size not in (128, 256):
            raise ValueError(f"Incorrect key size for CBC mode ({key_size} bits)")
        super().__init__(key, key_size, factory.block_size, iv_mode, iv_options, sector_size, iv_sector_size)

        if FAST_IV:
            self._cipher = AES.new(key[: self.key_size_bytes], AES.MODE_CBC, iv=b"\x00" * self.block_size)
            self._iv_view = get_iv_view(self._cipher, self.block_size)
        else:
            self._key = key[: self.key_size_bytes]

    def _crypt_sector(self, mode: int, buffer: bytearray, iv: bytes) -> None:
        if FAST_IV:
            self._iv_view[0 : self.block_size] = iv
            cipher = self._cipher
        else:
            cipher = AES.new(self._key, AES.MODE_CBC, iv=iv)

        (cipher.encrypt if mode == ENCRYPT else cipher.decrypt)(buffer, output=buffer)


class XtsMode(Cipher):
    """XTS mode implementation for FVE crypto."""

    def __init__(
        self,
        factory: AES,
        key: bytes,
        key_size: int,
        iv_mode: type[IV],
        iv_options: str,
        sector_size: int = 512,
        iv_sector_size: int = 512,
    ):
        if (len(key), key_size) not in ((32, 128), (64, 256)):
            raise ValueError(f"Incorrect key size for XTS mode ({len(key)} bytes, {key_size} bits)")
        super().__init__(key, key_size, key_size // 8, iv_mode, iv_options, sector_size, iv_sector_size)

        self._aes_cipher = factory.new(key[: self.block_size], factory.MODE_ECB)
        self._tweak_cipher = factory.new(key[self.block_size :], factory.MODE_ECB)

    def _crypt_sector(self, mode: int, buffer: bytearray, iv: bytes) -> None:
        tweak = self._tweak_cipher.encrypt(iv)
        _t = int.from_bytes(tweak, "little")

        crypt = self._aes_cipher.encrypt if mode == ENCRYPT else self._aes_cipher.decrypt

        view = buffer
        block_size = self.block_size

        for _ in range(self.sector_size // 16):
            block_slice = view[:16]
            xor(block_slice, tweak[:16], output=block_slice)
            crypt(block_slice, output=block_slice)
            xor(tweak[:16], block_slice, output=block_slice)

            _t <<= 1
            if _t & (1 << 128):
                _t ^= (1 << 128) | (0x87)
            tweak = (_t & ((1 << (block_size * 8)) - 1)).to_bytes(block_size, "little")

            view = view[16:]


class EBOIV(IV):
    """Encrypted byte-offset IV.

    Specific to Bitlocker.
    """

    def __init__(self, cipher: Cipher, key: bytes, iv_options: str | None = None):
        super().__init__(cipher, key)
        self._ecb_cipher = AES.new(key, AES.MODE_ECB)

    def generate(self, mode: int, iv: bytearray, data: bytearray, sector: int = 0) -> None:
        iv[:] = self._ecb_cipher.encrypt((sector * self.cipher.sector_size).to_bytes(16, "little"))


class ESSIV(IV):
    """Encrypted sector|salt IV.

    The sector number is encrypted with the bulk cipher using a salt as key. The salt should be
    derived from the bulk cipher's key via hashing.
    """

    def __init__(self, cipher: Cipher, key: bytes, iv_options: str | None = "sha256"):
        super().__init__(cipher, key)
        # Only support one cipher mode for now
        self._cipher = AES.new(hashlib.new(iv_options, key).digest(), AES.MODE_ECB)

    def generate(self, mode: int, iv: bytearray, data: bytearray, sector: int = 0) -> None:
        self._cipher.encrypt(sector.to_bytes(self.cipher.block_size, "little"), output=iv)


class Elephant(IV):
    """Extended eboiv with Elephant diffuser.

    Specific to Bitlocker. The key is always 64 bytes, but you need to take only the
    amount of bytes for the key size that you're working with.
    """

    def __init__(self, cipher: Cipher, key: bytes, iv_options: str | None = None):
        super().__init__(cipher, key)
        self._ecb_cipher = AES.new(key[32 : 32 + cipher.key_size_bytes], AES.MODE_ECB)
        self._eboiv = EBOIV(cipher, key[: cipher.key_size_bytes])

        self._sector_key = bytearray(32)
        self._sector_key_view = memoryview(self._sector_key)

    def _elephant(self, mode: int, data: bytearray, sector: int) -> None:
        sector_size = self.cipher.sector_size
        sector_key_view = self._sector_key_view

        # Generate the IV and sector key
        iv = bytearray((sector * sector_size).to_bytes(16, "little"))
        self._ecb_cipher.encrypt(iv, output=sector_key_view[:16])
        iv[15] = 0x80
        self._ecb_cipher.encrypt(iv, output=sector_key_view[16:])

        if mode == DECRYPT:
            # Apply diffuser B
            elephant.diffuser_b_decrypt(data, sector_size)

            # Apply diffuser A
            elephant.diffuser_a_decrypt(data, sector_size)

        # Apply sector key
        xor(data, self._sector_key * (sector_size // 32), output=data)

        if mode == ENCRYPT:
            # Apply diffuser A
            elephant.diffuser_a_encrypt(data, sector_size)

            # Apply diffuser B
            elephant.diffuser_b_encrypt(data, sector_size)

    def generate(self, mode: int, iv: bytearray, data: bytearray, sector: int = 0) -> None:
        if mode == ENCRYPT:
            self._elephant(mode, data, sector)
        self._eboiv.generate(mode, iv, data, sector)

    def post(self, mode: int, data: bytearray, sector: int = 0) -> None:
        if mode == DECRYPT:
            self._elephant(mode, data, sector)


def _create_cipher_factory(mode: type[Cipher]) -> Callable[..., Cipher]:
    def cipher_factory(factory: AES, **kwargs) -> Cipher:
        try:
            key = kwargs.pop("key")
            key_size = kwargs.pop("key_size")
            iv_mode = kwargs.pop("iv_mode")
        except KeyError as e:
            raise TypeError("Missing parameter:" + str(e))

        sector_size = kwargs.pop("sector_size", 512)
        iv_sector_size = kwargs.pop("iv_sector_size", 512)
        iv_options = kwargs.pop("iv_options", None)

        return mode(factory, key, key_size, iv_mode, iv_options, sector_size, iv_sector_size)

    return cipher_factory


_create_fve_ecb_cipher = _create_cipher_factory(EcbMode)
_create_fve_cbc_cipher = _create_cipher_factory(CbcMode)
_create_fve_xts_cipher = _create_cipher_factory(XtsMode)


def new(key: bytes, mode: int, *args, **kwargs) -> Cipher:
    """Create a new cipher object."""
    if mode not in ("ecb", "cbc", "xts"):
        return AES.new(key, mode, *args, **kwargs)

    if mode == "ecb":
        return _create_fve_ecb_cipher(AES, key=key, mode=mode, **kwargs)
    if mode == "cbc":
        return _create_fve_cbc_cipher(AES, key=key, mode=mode, **kwargs)
    if mode == "xts":
        return _create_fve_xts_cipher(AES, key=key, mode=mode, **kwargs)

    raise ValueError(f"Unsupported mode: {mode} (key: {key!r}, args: {args}, kwargs: {kwargs})")



from __future__ import annotations

import hashlib
import struct


def stretch(password: bytes, salt: bytes, rounds: int = 0x100000) -> bytes:
    """Stretch a password with a specified salt.

    Bitlocker uses this as the key derivation algorithm.
    """
    # Stretch data looks like the following:
    # chained hash | user hash | salt | counter
    # SHA256 digest length is 32 bytes
    # Salt length is 16 bytes
    # Counter is a uint64
    if len(password) != 32:
        raise ValueError("Invalid password length")

    if len(salt) != 16:
        raise ValueError("Invalid salt length")

    data = bytearray(32 + 32 + 16 + 8)
    view = memoryview(data)

    view[32:64] = password
    view[64:80] = salt

    for i in range(rounds):
        view[80:] = i.to_bytes(8, "little")
        view[:32] = hashlib.sha256(view).digest()

    return bytes(view[:32])


def derive_user_key(user_password: str) -> bytes:
    """Derive an AES key from a given user passphrase."""
    return hashlib.sha256(hashlib.sha256(user_password.encode("utf-16-le")).digest()).digest()


def derive_recovery_key(recovery_password: str) -> bytes:
    """Derive an AES key from a given recovery password."""
    check_recovery_password(recovery_password)

    blocks = recovery_password.split("-")
    key = b"".join(struct.pack("<H", int(block) // 11) for block in blocks)
    return hashlib.sha256(bytes(key)).digest()


def check_recovery_password(recovery_password: str) -> bool:
    """Check if a given recovery password is valid."""
    blocks = recovery_password.split("-")
    if len(blocks) != 8:
        raise ValueError("Invalid recovery password: invalid length")

    for block in blocks:
        if not block.isdigit():
            raise ValueError("Invalid recovery password: contains non-numeric value")

        value = int(block)
        if value % 11:
            raise ValueError("Invalid recovery password: block not divisible by 11")

        if value >= 2**16 * 11:
            raise ValueError("Invalid recovery password: larger than 2 ** 16 * 11 (720896)")

        digits = list(map(int, block))
        checksum = (digits[0] - digits[1] + digits[2] - digits[3] + digits[4]) % 11
        if checksum != digits[5]:
            raise ValueError("Invalid recovery password: invalid block checksum")

    return True



from __future__ import annotations

ENCRYPT = 0
DECRYPT = 1


class Cipher:
    def __init__(
        self,
        key: bytes,
        key_size: int,
        block_size: int,
        iv_mode: type[IV],
        iv_options: str,
        sector_size: int = 512,
        iv_sector_size: int = 512,
    ):
        self.key = key
        self.key_size = key_size
        self.key_size_bytes = key_size // 8
        self.block_size = block_size
        self.sector_size = sector_size
        self.iv_sector_size = iv_sector_size

        self.iv_mode = iv_mode(self, key, iv_options)

    def _crypt_sector(self, mode: int, buffer: bytearray, iv: bytes) -> None:
        raise NotImplementedError

    def _crypt(self, mode: int, ciphertext: bytes, sector: int = 0, output: bytearray | None = None) -> bytes | None:
        length = len(ciphertext)

        if length % self.block_size:
            raise ValueError("Ciphertext is not aligned to block size")

        out = output or bytearray(length)
        out[:] = ciphertext
        out_view = memoryview(out)

        iv = bytearray(self.iv_mode.iv_size)
        iv_view = memoryview(iv)

        iv_mode = self.iv_mode
        sector_size = self.sector_size
        sector_increment = sector_size // self.iv_sector_size

        for _ in range(length // sector_size):
            out_slice = out_view[:sector_size]

            # Generate the IV
            iv_mode.generate(mode, iv_view, out_slice, sector)

            # Do the crypting
            self._crypt_sector(mode, out_slice, iv)

            # Perform possible post operations for the IV
            iv_mode.post(mode, out_slice, sector)

            out_view = out_view[sector_size:]
            sector += sector_increment

        return None if output is not None else bytes(out)

    def encrypt(self, ciphertext: bytes, sector: int = 0, output: bytearray | None = None) -> bytes | None:
        return self._crypt(ENCRYPT, ciphertext, sector, output)

    def decrypt(self, ciphertext: bytes, sector: int = 0, output: bytearray | None = None) -> bytes | None:
        return self._crypt(DECRYPT, ciphertext, sector, output)


class IV:
    def __init__(self, cipher: Cipher, key: bytes, iv_options: str | None = None):
        self.cipher = cipher
        self.iv_size = cipher.block_size

    def generate(self, mode: int, iv: bytearray, data: bytearray, sector: int = 0) -> None:
        pass

    def post(self, mode: int, data: bytearray, sector: int = 0) -> None:
        pass


class Plain(IV):
    def generate(self, mode: int, iv: bytearray, data: bytearray, sector: int = 0) -> None:
        iv[:] = b"\x00" * self.iv_size
        iv[:4] = (sector & 0xFFFFFFFF).to_bytes(4, "little")


class Plain64(IV):
    def generate(self, mode: int, iv: bytearray, data: bytearray, sector: int = 0) -> None:
        iv[:] = b"\x00" * self.iv_size
        iv[:8] = sector.to_bytes(8, "little")


class Plain64BE(IV):
    def generate(self, mode: int, iv: bytearray, data: bytearray, sector: int = 0) -> None:
        iv[:] = b"\x00" * self.iv_size
        iv[:8] = sector.to_bytes(8, "big")


class EBOIV(IV):
    def __init__(self, cipher: Cipher, key: bytes, iv_options: str | None = None):
        # Implementation specific
        raise NotImplementedError


class ESSIV(IV):
    def __init__(self, cipher: Cipher, key: bytes, iv_options: str | None = None):
        # Implementation specific
        raise NotImplementedError


class Elephant(IV):
    def __init__(self, cipher: Cipher, key: bytes, iv_options: str | None = None):
        # Implementation specific
        raise NotImplementedError



from __future__ import annotations

import platform


# Reference: https://www.da.vidbuchanan.co.uk/blog/python-swar.html
# Sorry David
def xor_pseudo_simd(a: bytes, b: bytes, output: bytearray) -> None:
    output[:] = int.to_bytes(int.from_bytes(a, "little") ^ int.from_bytes(b, "little"), len(output), "little")


# On PyPy the naive loop is actually faster
# Also just use this as the default fallback, seems safer
def xor_naive(a: bytes, b: bytes, output: bytearray) -> None:
    for i in range(len(output)):
        output[i] = a[i] ^ b[i]


xor = xor_pseudo_simd if platform.python_implementation() == "CPython" else xor_naive



from __future__ import annotations

import base64
import json
from dataclasses import dataclass, field, fields
from types import UnionType
from typing import Any, get_args, get_origin, get_type_hints

from dissect.fve.luks.c_luks import c_luks


@dataclass
class JsonItem:
    _raw: dict | None = field(init=False, repr=False)

    @classmethod
    def from_json(cls, obj: str) -> JsonItem:  # Self, but that's >=3.11
        return cls.from_dict(json.loads(obj))

    @classmethod
    def from_dict(cls, obj: dict[str, str | int | dict | list]) -> JsonItem:  # Self, but that's >=3.11
        kwargs = {}
        raw = None
        type_info = get_type_hints(cls)
        for fld in fields(cls):
            if fld.name == "_raw":
                raw = obj
                continue

            value = obj.get(fld.name, None)
            kwargs[fld.name] = JsonItem._parse_type(type_info[fld.name], value)

        result = cls(**kwargs)
        result._raw = raw
        return result

    @staticmethod
    def _parse_type(type_: Any, value: str | int | dict | list) -> str | int | dict | list | bytes:
        result = None

        if type_ == type_ | None:
            result = JsonItem._parse_type(get_args(type_)[0], value) if value is not None else None
        elif get_origin(type_) is UnionType:
            for atype in get_args(type_):
                try:
                    result = JsonItem._parse_type(atype, value)
                    break
                except Exception:
                    continue
        elif get_origin(type_) is list:
            vtype = get_args(type_)[0]
            result = [JsonItem._parse_type(vtype, v) for v in value]
        elif get_origin(type_) is dict:
            ktype, vtype = get_args(type_)
            result = {JsonItem._parse_type(ktype, k): JsonItem._parse_type(vtype, v) for k, v in value.items()}
        elif type_ is bytes:
            result = base64.b64decode(value)
        elif issubclass(type_, JsonItem):
            result = type_.from_dict(value)
        else:
            result = type_(value)

        return result


@dataclass
class Config(JsonItem):
    json_size: int
    keyslots_size: int | None
    flags: list[str] | None
    requirements: list[str] | None


@dataclass
class KeyslotArea(JsonItem):
    type: str
    offset: int
    size: int
    # if type == "raw"
    encryption: str | None
    key_size: int | None
    # type == "datashift-checksum" has all the fields of "checksum" and "datashift"
    # if type == "checksum"
    hash: str | None
    sector_size: int | None
    # if type in ("datashift", "datashift-journal")
    shift_size: int | None


@dataclass
class KeyslotKdf(JsonItem):
    type: str
    salt: bytes
    # if type == "pbkdf2"
    hash: str | None
    iterations: int | None
    # if type in ("argon2i", "argin2id")
    time: int | None
    memory: int | None
    cpus: int | None


@dataclass
class KeyslotAf(JsonItem):
    type: str
    # if type == "luks1"
    stripes: int | None
    hash: str | None


@dataclass
class Keyslot(JsonItem):
    type: str
    key_size: int
    area: KeyslotArea
    priority: int | None
    # if type == "luks2"
    kdf: KeyslotKdf | None
    af: KeyslotAf | None
    # if type == "reencrypt"
    mode: str | None
    direction: str | None


@dataclass
class Digest(JsonItem):
    type: str
    keyslots: list[int]
    segments: list[int]
    salt: bytes
    digest: bytes
    # if type == "pbkdf2"
    hash: str | None
    iterations: int | None


@dataclass
class SegmentIntegrity(JsonItem):
    type: str
    journal_encryption: str
    journal_integrity: str


@dataclass
class Segment(JsonItem):
    type: str
    offset: int
    size: int | str
    flags: list[str] | None
    # if type == "crypt"
    iv_tweak: int | None
    encryption: str | None
    sector_size: int | None
    integrity: SegmentIntegrity | None


@dataclass
class Token(JsonItem):
    type: str
    keyslots: list[int]


@dataclass
class Metadata(JsonItem):
    config: Config
    keyslots: dict[int, Keyslot]
    digests: dict[int, Digest]
    segments: dict[int, Segment]
    tokens: dict[int, Token]

    @classmethod
    def from_luks1_header(self, header: c_luks.luks_phdr) -> Metadata:
        """Map LUKS1 header information into a :class:`Metadata` dataclass."""
        config = Config(0, None, None, None)
        keyslots = {}
        digests = {}
        segments = {}
        tokens = {}

        cipher_spec = "-".join(v.rstrip(b"\x00").decode() for v in [header.cipherName, header.cipherMode])
        hash_spec = header.hashSpec.rstrip(b"\x00").decode()

        for idx, block in enumerate(header.keyblock):
            if block.active == c_luks.LUKS_KEY_DISABLED:
                continue

            keyslots[idx] = Keyslot(
                type="luks1",
                key_size=header.keyBytes,
                area=KeyslotArea(
                    type="raw",
                    offset=block.keyMaterialOffset * 512,
                    size=header.keyBytes * block.stripes,
                    encryption=cipher_spec,
                    key_size=header.keyBytes,
                    hash=None,
                    sector_size=None,
                    shift_size=None,
                ),
                priority=None,
                kdf=KeyslotKdf(
                    type="pbkdf2",
                    salt=block.passwordSalt,
                    hash=hash_spec,
                    iterations=block.passwordIterations,
                    time=None,
                    memory=None,
                    cpus=None,
                ),
                af=KeyslotAf(type="luks1", stripes=block.stripes, hash=hash_spec),
                mode=None,
                direction=None,
            )

        digests[0] = Digest(
            type="pbkdf2",
            keyslots=list(keyslots.keys()),
            segments=[0],
            salt=header.mkDigestSalt,
            digest=header.mkDigest,
            hash=hash_spec,
            iterations=header.mkDigestIterations,
        )

        segments[0] = Segment(
            type="crypt",
            offset=header.payloadOffset * 512,
            size="dynamic",
            flags=None,
            iv_tweak=0,
            encryption=cipher_spec,
            sector_size=512,
            integrity=None,
        )

        return Metadata(config, keyslots, digests, segments, tokens)



from __future__ import annotations

import io
from typing import BinaryIO

from dissect.util.stream import AlignedStream

from dissect.fve.crypto import create_cipher


class CryptStream(AlignedStream):
    """Transparently decrypting stream.

    Args:
        fh: The original file-like object, usually the encrypted disk.
        cipher: The cipher name/specification.
        key: The encryption key.
        key_size: Optional key size hint.
        offset: Optional base offset to the encrypted region. Segment offset in LUKS.
        size: Optional size hint. If ``None`` or ``"dynamic"``, determine the size by seeking to the end of ``fh``.
        iv_tweak: Optional IV tweak, or offset.
        sector_size: Optional sector size. Defaults to 512.
    """

    def __init__(
        self,
        fh: BinaryIO,
        cipher: str,
        key: bytes,
        key_size: int | None = None,
        offset: int = 0,
        size: int | str | None = None,
        iv_tweak: int = 0,
        sector_size: int = 512,
    ):
        self.fh = fh
        self.cipher = create_cipher(cipher, key, key_size or len(key) * 8, sector_size, 512)
        self.offset = offset
        self.iv_tweak = iv_tweak
        self.sector_size = sector_size

        if size in (None, "dynamic"):
            size = fh.seek(0, io.SEEK_END) - offset

        super().__init__(size)

    def _read(self, offset: int, length: int) -> bytes:
        self.fh.seek(self.offset + offset)
        buf = bytearray(self.fh.read(length))
        self.cipher.decrypt(buf, (offset // 512) + self.iv_tweak, buf)
        return bytes(buf)



from __future__ import annotations

from typing import TYPE_CHECKING

# Only pycryptodome is supported right now
from dissect.fve.crypto import _pycryptodome
from dissect.fve.crypto.base import Plain, Plain64, Plain64BE

if TYPE_CHECKING:
    from dissect.fve.crypto.base import Cipher

IV_MODE_MAP = {
    "plain": Plain,
    "plain64": Plain64,
    "plain64be": Plain64BE,
    "eboiv": _pycryptodome.EBOIV,
    "essiv": _pycryptodome.ESSIV,
    "elephant": _pycryptodome.Elephant,
}


def create_cipher(
    spec: str, key: bytes, key_size: int | None = None, sector_size: int = 512, iv_sector_size: int = 512
) -> Cipher:
    """Create a cipher object according to a given cipher specification and key.

    For more information on the cipher specification, read the documentation on :func:`parse_cipher_spec`.

    Args:
        spec: The cipher specification to parse.
        key: The key to initialize the cipher with.
        key_size: Optional key size that overrides the specification key size.
        sector_size: Optional sector size.
    """
    cipher_name, cipher_mode, key_size, iv_name, iv_options = parse_cipher_spec(
        spec, key_size=key_size, key_size_hint=len(key) * 8
    )

    if cipher_name != "aes":
        raise ValueError("Only AES support is implemented")

    if cipher_mode not in ("ecb", "cbc", "xts"):
        raise ValueError(f"Invalid cipher mode: {cipher_name}-{cipher_mode} (from {spec})")

    if (iv := IV_MODE_MAP.get(iv_name)) is None:
        raise ValueError(f"Invalid iv mode: {iv_name}:{iv_options} (from {spec})")

    return _pycryptodome.new(
        key,
        cipher_mode,
        key_size=key_size,
        iv_mode=iv,
        iv_options=iv_options,
        sector_size=sector_size,
        iv_sector_size=iv_sector_size,
    )


def parse_cipher_spec(
    spec: str, key_size: int | None = None, key_size_hint: int | None = None
) -> tuple[str, str, int, str, str | None]:
    """Parse a cipher specification into a tuple of (cipher, mode, key size, iv mode, iv options).

    Inspired by and accepts LUKS/dm-crypt-like cipher specifications in the form of::

        cipher-mode-keysize-iv:ivopts

    The ``mode``, ``keysize``, ``iv`` and ``ivopts`` are optional and will default to ``cbc``,
    the ``key_size`` argument and ``plain`` respectively.

    Args:
        spec: The cipher specification to parse.
        key_size: Optional key size that overrides the specification key size.
        key_size_hint: Optional key size hint for the amount of bits that the key actually has.
    """
    cipher_name, _, tmp = spec.partition("-")
    cipher_mode, _, tmp = tmp.partition("-")

    result_key_size = key_size_hint
    specified_key_size, _, tmp = tmp.partition("-")
    if specified_key_size.isdigit():
        result_key_size = int(specified_key_size)
    else:
        if tmp:
            raise ValueError("Unexpected cipher spec format")
        tmp = specified_key_size

    if key_size:
        result_key_size = key_size

    if not result_key_size:
        raise ValueError("Missing key size")

    iv_name = None
    iv_options = None
    iv_name, _, iv_options = tmp.partition(":")

    if not cipher_mode:
        cipher_mode = "cbc"

    if not iv_name:
        iv_name = "plain"
        iv_options = None

    if cipher_mode == "xts" and key_size_hint == result_key_size:
        result_key_size //= 2

    return cipher_name, cipher_mode, result_key_size, iv_name, iv_options or None



from __future__ import annotations

try:
    from dissect.fve import _native

    HAS_ARGON2 = True

    hash_secret_raw = _native.crypto.argon2.hash_secret_raw

except (ImportError, AttributeError):
    try:
        import argon2

        HAS_ARGON2 = True

        def hash_secret_raw(
            secret: bytes, salt: bytes, time_cost: int, memory_cost: int, parallelism: int, hash_len: int, type: str
        ) -> bytes:
            return argon2.low_level.hash_secret_raw(
                secret,
                salt,
                time_cost,
                memory_cost,
                parallelism,
                hash_len,
                {"argon2i": argon2.low_level.Type.I, "argon2id": argon2.low_level.Type.ID}[type],
            )
    except ImportError:
        HAS_ARGON2 = False

        def hash_secret_raw(
            secret: bytes, salt: bytes, time_cost: int, memory_cost: int, parallelism: int, hash_len: int, type: str
        ) -> bytes:
            raise RuntimeError(
                "A native Argon2 implementation is not available. "
                "Please ensure that the dissect.fve._native module is built and accessible, "
                "or install `argon2-cffi` from PyPI."
            )
